@@ -2,14 +2,16 @@
 PC-BASIC - lister.py
 Convert tokenised to plain-text format
 
-(c) 2013--2018 Rob Hagemans
+(c) 2013--2019 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
-import string
 import struct
 
+from ...compat import int2byte
+
 from ..base import tokens as tk
+from ..base.tokens import DIGITS, ALPHANUMERIC
 from .. import values
 
 
@@ -27,15 +29,15 @@ class Lister(object):
         if current_line < 0:
             # detokenise_line_number has returned -1 and left us at: .. 00 | _00_ 00 1A
             # stream ends or end of file sequence \x00\x00\x1A
-            return -1, '', 0
-        elif current_line == 0 and ins.peek() == ' ':
+            return -1, b'', 0
+        elif current_line == 0 and ins.peek() == b' ':
             # ignore up to one space after line number 0
             ins.read(1)
-        linum = bytearray(str(current_line))
+        linum = bytearray(b'%d' % (current_line,))
         # write one extra whitespace character after line number
         # unless first char is TAB
-        if ins.peek() != '\t':
-            linum += bytearray(' ')
+        if ins.peek() != b'\t':
+            linum += bytearray(b' ')
         line, textpos = self.detokenise_compound_statement(ins, bytepos)
         return current_line, linum + line, textpos + len(linum) + 1
 
@@ -50,7 +52,7 @@ class Lister(object):
 
     def token_to_line_number(self, trail):
         """Unpack a line number token trail, -1 if end of program."""
-        if len(trail) < 4 or trail[:2] == '\0\0':
+        if len(trail) < 4 or trail[:2] == b'\0\0':
             return -1
         return struct.unpack_from('<H', trail, 2)[0]
 
@@ -61,14 +63,14 @@ class Lister(object):
         output = bytearray()
         while True:
             s = ins.read(1)
-            if not textpos and ins.tell() >= bytepos:
+            if not textpos and bytepos is not None and ins.tell() >= bytepos:
                 textpos = len(output)
             if s in tk.END_LINE:
                 # \x00 ends lines and comments when listed,
                 # if not inside a number constant
                 # stream ended or end of line
                 break
-            elif s == '"':
+            elif s == b'"':
                 # start of literal string, passed verbatim
                 # until a closing quote or EOL comes by
                 # however number codes are *printed* as the corresponding numbers,
@@ -77,13 +79,13 @@ class Lister(object):
                 litstring = not litstring
             elif s in tk.NUMBER or s in tk.LINE_NUMBER:
                 self._detokenise_number(ins, s, output)
-            elif comment or litstring or ('\x20' <= s <= '\x7E'):
+            elif comment or litstring or (b'\x20' <= s <= b'\x7E'):
                 # honest ASCII
                 output += s
-            elif s == '\x0A':
+            elif s == b'\x0A':
                 # LF becomes LF CR
-                output += '\x0A\x0D'
-            elif s <= '\x09':
+                output += b'\x0A\x0D'
+            elif s <= b'\x09':
                 # controls that do not double as tokens
                 output += s
             else:
@@ -104,24 +106,27 @@ class Lister(object):
                 keyword = self._token_to_keyword[s]
                 ins.read(1)
             except KeyError:
-                output += s[0]
+                output += s[:1]
                 return False
         # when we're here, s is an actual keyword token.
         # letter or number followed by token is separated by a space
-        if (output and chr(output[-1]) in (string.digits + string.ascii_letters) and s not in tk.OPERATOR):
-            output += ' '
+        if (
+                output and int2byte(output[-1]) in ALPHANUMERIC
+                and s not in tk.OPERATOR
+            ):
+            output += b' '
         output += keyword
         comment = False
-        if keyword == "'":
+        if keyword == b"'":
             comment = True
         elif keyword == tk.KW_REM:
             nxt = ins.read(1)
-            if nxt == '':
+            if nxt == b'':
                 pass
             elif nxt == tk.O_REM: # '
                 # if next char is token('), we have the special value REM'
                 # -- replaced by ' below.
-                output += "'"
+                output += b"'"
             else:
                 # otherwise, it's part of the comment or an EOL or whatever,
                 # pass back to stream so it can be processed
@@ -129,31 +134,36 @@ class Lister(object):
             comment = True
         # check for special cases
         #   [:REM']   ->  [']
-        if len(output) > 4 and str(output[-5:]) == ":REM'":
-            output[:] = output[:-5] + "'"
+        if len(output) > 4 and bytes(output[-5:]) == b":REM'":
+            output[:] = output[:-5] + b"'"
         #   [WHILE+]  ->  [WHILE]
-        elif len(output) > 5 and str(output[-6:]) == 'WHILE+':
+        elif len(output) > 5 and bytes(output[-6:]) == b'WHILE+':
             output[:] = output[:-1]
         #   [:ELSE]  ->  [ELSE]
         # note that anything before ELSE gets cut off,
         # e.g. if we have 1ELSE instead of :ELSE it also becomes ELSE
         # SIC: len(output) > 4 and str(output[-4:])
-        elif len(output) > 4 and str(output[-4:]) == tk.KW_ELSE:
-            if (len(output) > 5 and chr(output[-5]) == ':' and
-                        chr(output[-6]) in string.digits):
-                output[:] = output[:-5] + ' ' + tk.KW_ELSE
+        elif len(output) > 4 and bytes(output[-4:]) == tk.KW_ELSE:
+            if (
+                    len(output) > 5 and int2byte(output[-5]) == b':' and
+                    int2byte(output[-6]) in DIGITS
+                ):
+                output[:] = output[:-5] + b' ' + tk.KW_ELSE
             else:
                 output[:] = output[:-5] + tk.KW_ELSE
         # token followed by token or number is separated by a space,
         # except operator tokens and SPC(, TAB(, FN, USR
         nxt = ins.peek()
-        if (not comment and
-                nxt not in tk.END_LINE + tk.OPERATOR +
-                        (tk.O_REM, '"', ',', ';', ' ', ':', '(', ')', '$',
-                         '%', '!', '#', '_', '@', '~', '|', '`') and
-                s not in tk.OPERATOR + (tk.TAB, tk.SPC, tk.USR, tk.FN)):
+        if (
+                not comment
+                and nxt not in tk.END_LINE + tk.OPERATOR + (
+                    tk.O_REM, b'"', b',', b';', b' ', b':', b'(', b')', b'$',
+                    b'%', b'!', b'#', b'_', b'@', b'~', b'|', b'`'
+                )
+                and s not in tk.OPERATOR + (tk.TAB, tk.SPC, tk.USR, tk.FN)
+            ):
             # excluding TAB( SPC( and FN. \xD9 is ', \xD1 is FN, \xD0 is USR.
-            output += ' '
+            output += b' '
         return comment
 
     def _detokenise_number(self, ins, lead, output):
@@ -168,13 +178,13 @@ class Lister(object):
         elif lead == tk.T_HEX:
             output += b'&H' + self._values.from_bytes(trail).to_hex()
         elif lead == tk.T_BYTE:
-            output += str(ord(trail))
+            output += b'%d' % (ord(trail),)
         elif tk.C_0 <= lead <= tk.C_10:
-            output += str(ord(lead) - ord(tk.C_0))
+            output += b'%d' % (ord(lead) - ord(tk.C_0),)
         elif lead in tk.LINE_NUMBER:
             # 0D: line pointer (unsigned int) - this token should not be here;
             #     interpret as line number and carry on
             # 0E: line number (unsigned int)
-            output += str(struct.unpack(b'<H', trail)[0])
+            output += b'%d' % (struct.unpack('<H', trail)[0],)
         elif lead in (tk.T_SINGLE, tk.T_DOUBLE, tk.T_INT):
             output += self._values.from_bytes(trail).to_str(leading_space=False, type_sign=True)
