@@ -2,16 +2,18 @@
 PC-BASIC - files.py
 Devices, Files and I/O operations
 
-(c) 2013--2018 Rob Hagemans
+(c) 2013--2019 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
 import os
 import sys
-import string
 import logging
 import io
 
+from ...compat import xrange, int2byte, text_type
+from ...compat import iterchar, getcwdu
+
 from ..base import error
 from ..base import tokens as tk
 from .. import values
@@ -27,7 +29,10 @@ from . import parports
 DOS_DEVICE_FILES = (b'AUX', b'CON', b'NUL', b'PRN')
 
 # default mount dictionary
-DEFAULT_MOUNTS = {b'Z': (os.getcwdu(), u'')}
+DEFAULT_MOUNTS = {b'Z': (getcwdu(), u'')}
+
+# allowable drive letters in GW-BASIC are letters or @
+DRIVE_LETTERS = b'@' + tk.UPPERCASE
 
 
 ############################################################################
@@ -40,20 +45,21 @@ class Files(object):
             self, values, memory, queues, keyboard, display,
             max_files, max_reclen, serial_buffer_size,
             device_params, current_device, mount_dict,
-            utf8, universal):
+            text_mode, soft_linefeed
+        ):
         """Initialise files."""
         # for wait() in files_
         self._queues = queues
         self._values = values
         self._memory = memory
-        self._fields = self._memory.fields
         self.files = {}
         self.max_files = max_files
         self.max_reclen = max_reclen
         self._init_devices(
-                values, queues, display, keyboard,
-                device_params, current_device, mount_dict,
-                serial_buffer_size, utf8, universal)
+            values, queues, display, keyboard,
+            device_params, current_device, mount_dict,
+            serial_buffer_size, text_mode, soft_linefeed
+        )
 
     ###########################################################################
     # file management
@@ -72,8 +78,10 @@ class Files(object):
             f.close()
         self.files = {}
 
-    def open(self, number, description, filetype, mode=b'I', access=b'', lock=b'',
-                reclen=128, seg=0, offset=0, length=0):
+    def open(
+            self, number, description, filetype, mode=b'I', access=b'', lock=b'',
+            reclen=128, seg=0, offset=0, length=0
+        ):
         """Open a file on a device specified by description."""
         if (not description) or (number < 0) or (number > self.max_files):
             # bad file number; also for name='', for some reason
@@ -83,11 +91,12 @@ class Files(object):
         mode = mode.upper()
         device, dev_param = self._get_device_param(description, mode)
         # get the field buffer
-        field = self._fields[number] if number else None
+        field = self._memory.fields[number] if number else None
         # open the file on the device
         new_file = device.open(
-                number, dev_param, filetype, mode, access, lock,
-                reclen, seg, offset, length, field)
+            number, dev_param, filetype, mode, access, lock,
+            reclen, seg, offset, length, field
+        )
         logging.debug(
             'Opened file %r as #%d (type %s, mode %s)', dev_param, number, filetype, mode
         )
@@ -117,9 +126,10 @@ class Files(object):
     # device management
 
     def _init_devices(
-                self, values, queues, display, keyboard,
-                device_params, current_device, mount_dict,
-                serial_in_size, utf8, universal):
+            self, values, queues, display, keyboard,
+            device_params, current_device, mount_dict,
+            serial_in_size, text_mode, soft_linefeed
+        ):
         """Initialise devices."""
         # screen device, for files_()
         self._screen = display.text_screen
@@ -136,7 +146,8 @@ class Files(object):
             b'COM2:': ports.COMDevice(device_params.get(b'COM2:', None), queues, serial_in_size),
             # parallel devices - LPT1: must always be available
             b'LPT1:': parports.LPTDevice(
-                        device_params.get(b'LPT1:', None), devicebase.nullstream(), codepage),
+                device_params.get(b'LPT1:', None), devicebase.nullstream(), codepage
+            ),
             b'LPT2:': parports.LPTDevice(device_params.get(b'LPT2:', None), None, codepage),
             b'LPT3:': parports.LPTDevice(device_params.get(b'LPT3:', None), None, codepage),
         }
@@ -145,7 +156,7 @@ class Files(object):
         self.kybd_file = self._devices[b'KYBD:'].device_file
         self.lpt1_file = self._devices[b'LPT1:'].device_file
         # disks
-        self._init_disk_devices(mount_dict, current_device, codepage, utf8, universal)
+        self._init_disk_devices(mount_dict, current_device, codepage, text_mode, soft_linefeed)
 
     def close_devices(self):
         """Close device master files."""
@@ -180,15 +191,15 @@ class Files(object):
             # MS-DOS device aliases - these can't be names of disk files
             if device != self._devices[b'CAS1:'] and name in DOS_DEVICE_FILES:
                 if name == b'AUX':
-                    device, dev_param = self._devices[b'COM1:'], ''
+                    device, dev_param = self._devices[b'COM1:'], b''
                 elif name == b'CON' and mode == b'I':
-                    device, dev_param = self._devices[b'KYBD:'], ''
+                    device, dev_param = self._devices[b'KYBD:'], b''
                 elif name == b'CON' and mode == b'O':
-                    device, dev_param = self._devices[b'SCRN:'], ''
+                    device, dev_param = self._devices[b'SCRN:'], b''
                 elif name == b'PRN':
-                    device, dev_param = self._devices[b'LPT1:'], ''
+                    device, dev_param = self._devices[b'LPT1:'], b''
                 elif name == b'NUL':
-                    device, dev_param = devicebase.NullDevice(), ''
+                    device, dev_param = devicebase.NullDevice(), b''
             else:
                 # open file on default device
                 dev_param = name
@@ -271,7 +282,7 @@ class Files(object):
                 error.range_check(0, 255, width)
                 name, index = next(args)
                 name = self._memory.complete_name(name)
-                self._fields[number].attach_var(name, index, offset, width)
+                self._memory.fields[number].attach_var(name, index, offset, width)
                 offset += width
         except StopIteration:
             pass
@@ -359,21 +370,21 @@ class Files(object):
         else:
             file_number = values.to_int(file_number)
             error.range_check(0, 255, file_number)
-            output = self.get(file_number, 'OAR')
+            output = self.get(file_number, b'OAR')
         outstrs = []
         try:
             while True:
                 expr = next(args)
                 if isinstance(expr, values.String):
-                    outstrs.append('"%s"' % expr.to_str())
+                    outstrs.append(b'"%s"' % expr.to_str())
                 else:
                     outstrs.append(values.to_repr(expr, leading_space=False, type_sign=False))
         except StopIteration:
             # write the whole thing as one thing (this affects line breaks)
-            output.write_line(','.join(outstrs))
+            output.write_line(b','.join(outstrs))
         except error.BASICError:
             if outstrs:
-                output.write(','.join(outstrs) + ',')
+                output.write(b','.join(outstrs) + b',')
             raise
 
     def width_(self, args):
@@ -386,7 +397,7 @@ class Files(object):
         elif isinstance(file_or_device, values.Number):
             file_or_device = values.to_int(file_or_device)
             error.range_check(0, 255, file_or_device)
-            dev = self.get(file_or_device, mode='IOAR')
+            dev = self.get(file_or_device, mode=b'IOAR')
             w = values.to_int(next(args))
         else:
             expr = next(args)
@@ -418,7 +429,7 @@ class Files(object):
         if file_number is not None:
             file_number = values.to_int(file_number)
             error.range_check(0, 255, file_number)
-            output = self.get(file_number, 'OAR')
+            output = self.get(file_number, b'OAR')
             screen = None
         else:
             # neither LPRINT not a file number: print to screen
@@ -439,7 +450,7 @@ class Files(object):
         thefile = self.get(num)
         control_string = values.next_string(args)
         list(args)
-        logging.warning("IOCTL statement not implemented.")
+        logging.warning('IOCTL statement not implemented.')
         raise error.BASICError(error.IFC)
 
     def motor_(self, args):
@@ -473,7 +484,7 @@ class Files(object):
         num, = args
         num = values.to_integer(num)
         eof = self._values.new_integer()
-        if not num.is_zero() and self._get_from_integer(num, 'IR').eof():
+        if not num.is_zero() and self._get_from_integer(num, b'IR').eof():
             eof = eof.from_int(-1)
         return eof
 
@@ -489,7 +500,7 @@ class Files(object):
         num, = args
         num = values.to_int(num)
         error.range_check(0, 3, num)
-        printer = self._devices['LPT%d:' % max(1, num)]
+        printer = self._devices[b'LPT%d:' % max(1, num)]
         col = printer.device_settings.col
         # follow weird GW-BASIC behaviour
         # this is reported as 1 if it equals the DEVICE's width plus one
@@ -507,7 +518,7 @@ class Files(object):
             filenum = values.to_int(filenum)
             error.range_check(0, 255, filenum)
             # raise BAD FILE MODE (not BAD FILE NUMBER) if the file is not open
-            file_obj = self.get(filenum, mode='IR', not_open=error.BAD_FILE_MODE)
+            file_obj = self.get(filenum, mode=b'IR', not_open=error.BAD_FILE_MODE)
         else:
             file_obj = self.kybd_file
         list(args)
@@ -528,7 +539,7 @@ class Files(object):
         # raise BAD FILE NUMBER if the file is not open
         infile = self.get(num)
         list(args)
-        logging.warning("IOCTL$ function not implemented.")
+        logging.warning('IOCTL$ function not implemented.')
         raise error.BASICError(error.IFC)
 
     def erdev_(self, args):
@@ -555,18 +566,16 @@ class Files(object):
     ###########################################################################
     # disk devices
 
-    # allowable drive letters in GW-BASIC are letters or @
-    drive_letters = b'@' + string.ascii_uppercase
-
     def _init_disk_devices(
             self, mount_dict, current_device,
-            codepage, utf8, universal):
+            codepage, text_mode, soft_linefeed
+        ):
         """Initialise disk devices."""
         # use None to request default mounts, use {} for no mounts
         if mount_dict is None:
             mount_dict = DEFAULT_MOUNTS
         # disk devices
-        for letter in self.drive_letters:
+        for letter in iterchar(DRIVE_LETTERS):
             if not mount_dict:
                 mount_dict = {}
             if letter in mount_dict:
@@ -575,9 +584,11 @@ class Files(object):
                 path, cwd = None, u''
             # treat device @: separately - internal disk
             disk_class = disk.InternalDiskDevice if letter == b'@' else disk.DiskDevice
-            self._devices[letter + b':'] = disk_class(letter, path, cwd, codepage, utf8, universal)
-        # allow upper or lower case, unicode or str, with or without :
-        if isinstance(current_device, unicode):
+            self._devices[letter + b':'] = disk_class(
+                letter, path, cwd, codepage, text_mode, soft_linefeed
+            )
+        # allow upper or lower case, unicode or bytes, with or without :
+        if isinstance(current_device, text_type):
             current_device = current_device.encode('ascii')
         self._current_device = current_device.split(b':')[0].upper()
 
@@ -595,14 +606,14 @@ class Files(object):
             except KeyError:
                 raise error.BASICError(error.DEVICE_UNAVAILABLE)
         # must be a disk device
-        if dev not in self.drive_letters:
+        if dev not in DRIVE_LETTERS:
             raise error.BASICError(error.DEVICE_UNAVAILABLE)
         return self._devices[dev + b':'], spec
 
     def get_native_cwd(self):
         """Get current working directory on current drive."""
         # must be a disk device
-        if self._current_device not in self.drive_letters:
+        if self._current_device not in DRIVE_LETTERS:
             raise error.BASICError(error.IFC)
         return self._devices[self._current_device + b':'].get_native_cwd()
 
@@ -669,7 +680,7 @@ class Files(object):
         dev, path = self._get_diskdevice_and_path(pathmask)
         # retrieve files first (to ensure correct path/file not found errors)
         output = dev.listdir(path)
-        num_cols = self._screen.mode.width//20
+        num_cols = self._screen.mode.width // 20
         # output working dir in DOS format
         # NOTE: this is always the current dir, not the one being listed
         self._screen.write_line(dev.get_cwd())
@@ -678,7 +689,7 @@ class Files(object):
         # output files
         for i, cols in enumerate(output[j:j+num_cols] for j in xrange(0, len(output), num_cols)):
             self._screen.write_line(b' '.join(cols))
-            if not (i%4):
+            if not (i % 4):
                 # allow to break during dir listing & show names flowing on screen
                 self._queues.wait()
             i += 1
