@@ -2,12 +2,14 @@
 PC-BASIC - textscreen.py
 Text operations
 
-(c) 2013--2018 Rob Hagemans
+(c) 2013--2019 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
 import logging
 
+from ...compat import iterchar, iteritems, int2byte
+
 from ..base import signals
 from ..base import error
 from ..base import tokens as tk
@@ -43,8 +45,9 @@ class TextScreen(object):
         if not fonts:
             fonts = {8: {}}
         self.fonts = {
-                height: font.Font(height, font_dict)
-                for height, font_dict in fonts.iteritems()}
+            height: font.Font(height, font_dict)
+            for height, font_dict in iteritems(fonts)
+        }
         # function key macros
         self.bottom_bar = BottomBar()
 
@@ -57,8 +60,10 @@ class TextScreen(object):
         # set up glyph cache and preload halfwidth glyphs (i.e. single-byte code points)
         self._glyphs = font.GlyphCache(self.mode, self.fonts, self.codepage, self.queues)
         # build the screen buffer
-        self.text = TextBuffer(self.attr, self.mode.width, self.mode.height, self.mode.num_pages,
-                               self.codepage, do_fullwidth=(self.mode.font_height >= 14))
+        self.text = TextBuffer(
+            self.attr, self.mode.width, self.mode.height, self.mode.num_pages,
+            self.codepage, do_fullwidth=(self.mode.font_height >= 14)
+        )
         # pixel buffer
         self.pixels = pixels
         # redraw key line
@@ -81,8 +86,10 @@ class TextScreen(object):
     def check_font_available(self, mode):
         """Raise IFC if no suitable font available for this mode."""
         if mode.font_height not in self.fonts:
-            logging.warning('No %d-pixel font available. Could not enter video mode %s.',
-                            mode.font_height, mode.name)
+            logging.warning(
+                'No %d-pixel font available. Could not enter video mode %s.',
+                mode.font_height, mode.name
+            )
             raise error.BASICError(error.IFC)
 
     def rebuild(self):
@@ -90,11 +97,13 @@ class TextScreen(object):
         # send the glyph dict to interface if necessary
         self._glyphs.submit()
         # fix the cursor
-        self.queues.video.put(signals.Event(signals.VIDEO_SET_CURSOR_SHAPE,
-                (self.cursor.width, self.mode.font_height,
-                 self.cursor.from_line, self.cursor.to_line)))
-        self.queues.video.put(signals.Event(signals.VIDEO_MOVE_CURSOR,
-                (self.current_row, self.current_col)))
+        self.queues.video.put(signals.Event(
+            signals.VIDEO_SET_CURSOR_SHAPE,
+            (self.cursor.width, self.mode.font_height, self.cursor.from_line, self.cursor.to_line)
+        ))
+        self.queues.video.put(signals.Event(
+            signals.VIDEO_MOVE_CURSOR, (self.current_row, self.current_col)
+        ))
         if self.mode.is_text_mode:
             attr = self.text.get_attr(self.apagenum, self.current_row, self.current_col)
             fore, _, _, _ = self.mode.split_attr(attr & 0xf)
@@ -108,13 +117,16 @@ class TextScreen(object):
                 self.refresh_range(pagenum, row+1, 1, self.mode.width, text_only=True)
             # redraw graphics
             if not self.mode.is_text_mode:
-                self.queues.video.put(signals.Event(signals.VIDEO_PUT_RECT, (pagenum, 0, 0,
-                                self.mode.pixel_width-1, self.mode.pixel_height-1,
-                                self.pixels.pages[pagenum].buffer)))
+                self.queues.video.put(signals.Event(
+                    signals.VIDEO_PUT_RECT, (
+                        pagenum, 0, 0, self.mode.pixel_width-1, self.mode.pixel_height-1,
+                        self.pixels.pages[pagenum].buffer
+                    )
+                ))
 
-    def __str__(self):
-        """Return a string representation of the screen buffer (for debugging)."""
-        return str(self.text)
+    def __repr__(self):
+        """Return an ascii representation of the screen buffer (for debugging)."""
+        return repr(self.text)
 
     ##########################################################################
 
@@ -122,47 +134,47 @@ class TextScreen(object):
         """Write a string to the screen at the current position."""
         if do_echo:
             # CR -> CRLF, CRLF -> CRLF LF
-            self._io_streams.write(''.join([ ('\r\n' if c == '\r' else c) for c in s ]))
-        last = ''
+            self._io_streams.write(b''.join([(b'\r\n' if c == b'\r' else c) for c in iterchar(s)]))
+        last = b''
         # if our line wrapped at the end before, it doesn't anymore
         self.text.pages[self.apagenum].row[self.current_row-1].wrap = False
-        for c in s:
+        for c in iterchar(s):
             row, col = self.current_row, self.current_col
-            if c == '\t':
+            if c == b'\t':
                 # TAB
-                num = (8 - (col - 1 - 8 * int((col-1) / 8)))
+                num = (8 - (col - 1 - 8 * int((col-1) // 8)))
                 for _ in range(num):
-                    self.write_char(' ')
-            elif c == '\n':
+                    self.write_char(b' ')
+            elif c == b'\n':
                 # LF
                 # exclude CR/LF
-                if last != '\r':
+                if last != b'\r':
                     # LF connects lines like word wrap
                     self.text.pages[self.apagenum].row[row-1].wrap = True
                     self.set_pos(row + 1, 1, scroll_ok)
-            elif c == '\r':
+            elif c == b'\r':
                 # CR
                 self.text.pages[self.apagenum].row[row-1].wrap = False
                 self.set_pos(row + 1, 1, scroll_ok)
-            elif c == '\a':
+            elif c == b'\a':
                 # BEL
                 self.sound.beep()
-            elif c == '\x0B':
+            elif c == b'\x0B':
                 # HOME
                 self.set_pos(1, 1, scroll_ok)
-            elif c == '\x0C':
+            elif c == b'\x0C':
                 # CLS
                 self.clear_view()
-            elif c == '\x1C':
+            elif c == b'\x1C':
                 # RIGHT
                 self.set_pos(row, col + 1, scroll_ok)
-            elif c == '\x1D':
+            elif c == b'\x1D':
                 # LEFT
                 self.set_pos(row, col - 1, scroll_ok)
-            elif c == '\x1E':
+            elif c == b'\x1E':
                 # UP
                 self.set_pos(row - 1, col, scroll_ok)
-            elif c == '\x1F':
+            elif c == b'\x1F':
                 # DOWN
                 self.set_pos(row + 1, col, scroll_ok)
             else:
@@ -183,7 +195,7 @@ class TextScreen(object):
         for i, l in enumerate(cuts):
             # clear_line looks back along wraps, use screen.clear_from instead
             self.clear_from(self.current_row, 1)
-            self.write(str(l))
+            self.write(l)
             if i != len(cuts) - 1:
                 self.write(b'\n')
         if newline:
@@ -234,7 +246,7 @@ class TextScreen(object):
     def start_line(self):
         """Move the cursor to the start of the next line, this line if empty."""
         if self.current_col != 1:
-            self._io_streams.write('\r\n')
+            self._io_streams.write(b'\r\n')
             self._check_pos(scroll_ok=True)
             self.set_pos(self.current_row + 1, 1)
         # ensure line above doesn't wrap
@@ -288,7 +300,7 @@ class TextScreen(object):
         # see if we need to move to the next row
         if self.current_col > self.mode.width:
             if self.current_row < self.scroll_area.bottom or scroll_ok:
-                # either we don't nee to scroll, or we're allowed to
+                # either we don't need to scroll, or we're allowed to
                 self.current_col -= self.mode.width
                 self.current_row += 1
             else:
@@ -310,8 +322,7 @@ class TextScreen(object):
             self.current_row = self.scroll_area.top
         self._move_cursor(self.current_row, self.current_col)
         # signal position change
-        return (self.current_row == oldrow and
-                 self.current_col == oldcol)
+        return (self.current_row == oldrow and self.current_col == oldcol)
 
     def _move_cursor(self, row, col):
         """Move the cursor to a new position."""
@@ -320,11 +331,13 @@ class TextScreen(object):
         width = self.text.get_charwidth(self.apagenum, self.current_row, self.current_col)
         self.cursor.set_width(width)
         # set the cursor's attribute to that of the current location
-        fore, _, _, _ = self.mode.split_attr(0xf &
-                self.text.get_attr(self.apagenum, self.current_row, self.current_col))
+        fore, _, _, _ = self.mode.split_attr(
+            0xf & self.text.get_attr(self.apagenum, self.current_row, self.current_col)
+        )
         self.cursor.reset_attr(fore)
-        self.queues.video.put(signals.Event(signals.VIDEO_MOVE_CURSOR,
-                (self.current_row, self.current_col)))
+        self.queues.video.put(signals.Event(
+            signals.VIDEO_MOVE_CURSOR, (self.current_row, self.current_col))
+        )
 
     def move_to_end(self):
         """Jump to end of logical line; follow wraps (END)."""
@@ -360,16 +373,19 @@ class TextScreen(object):
             # ensure glyph is stored
             self._glyphs.check_char(char)
             fore, back, blink, underline = self.mode.split_attr(attr)
-            self.queues.video.put(signals.Event(signals.VIDEO_PUT_GLYPH, (
+            self.queues.video.put(signals.Event(
+                signals.VIDEO_PUT_GLYPH, (
                     pagenum, r, c, self.codepage.to_unicode(char, u'\0'),
                     len(char) > 1, fore, back, blink, underline,
-            )))
+                )
+            ))
             if not self.mode.is_text_mode and not text_only:
                 # update pixel buffer
                 x0, y0, x1, y1, sprite = self._glyphs.get_sprite(r, c, char, fore, back)
                 self.pixels.pages[self.apagenum].put_rect(x0, y0, x1, y1, sprite, tk.PSET)
                 self.queues.video.put(signals.Event(
-                        signals.VIDEO_PUT_RECT, (self.apagenum, x0, y0, x1, y1, sprite)))
+                    signals.VIDEO_PUT_RECT, (self.apagenum, x0, y0, x1, y1, sprite)
+                ))
 
     def _redraw_row(self, start, row, wrap=True):
         """Draw the screen row, wrapping around and reconstructing DBCS buffer."""
@@ -377,9 +393,12 @@ class TextScreen(object):
             for i in range(start, self.text.pages[self.apagenum].row[row-1].end):
                 # redrawing changes colour attributes to current foreground (cf. GW)
                 # don't update all dbcs chars behind at each put
-                char = chr(self.text.get_char(self.apagenum, row, i+1))
+                char = int2byte(self.text.get_char(self.apagenum, row, i+1))
                 self.put_char_attr(self.apagenum, row, i+1, char, self.attr, one_only=True)
-            if (wrap and self.text.pages[self.apagenum].row[row-1].wrap and row >= 0 and row < self.text.height-1):
+            if (
+                    wrap and self.text.pages[self.apagenum].row[row-1].wrap and
+                    row >= 0 and row < self.text.height-1
+                ):
                 row += 1
                 start = 0
             else:
@@ -463,10 +482,11 @@ class TextScreen(object):
         if from_line is None:
             from_line = self.scroll_area.top
         _, back, _, _ = self.mode.split_attr(self.attr)
-        self.queues.video.put(signals.Event(signals.VIDEO_SCROLL_UP,
-                    (from_line, self.scroll_area.bottom, back)))
+        self.queues.video.put(signals.Event(
+            signals.VIDEO_SCROLL_UP, (from_line, self.scroll_area.bottom, back)
+        ))
         if self.current_row > from_line:
-            self.current_row -= 1
+            self._move_cursor(self.current_row - 1, self.current_col)
         # sync buffers with the new screen reality:
         self.text.scroll_up(self.apagenum, from_line, self.scroll_area.bottom, self.attr)
         if not self.mode.is_text_mode:
@@ -479,10 +499,11 @@ class TextScreen(object):
     def scroll_down(self, from_line):
         """Scroll the scroll region down by one line, starting at from_line."""
         _, back, _, _ = self.mode.split_attr(self.attr)
-        self.queues.video.put(signals.Event(signals.VIDEO_SCROLL_DOWN,
-                    (from_line, self.scroll_area.bottom, back)))
+        self.queues.video.put(signals.Event(
+            signals.VIDEO_SCROLL_DOWN, (from_line, self.scroll_area.bottom, back)
+        ))
         if self.current_row >= from_line:
-            self.current_row += 1
+            self._move_cursor(self.current_row + 1, self.current_col)
         # sync buffers with the new screen reality:
         self.text.scroll_down(self.apagenum, from_line, self.scroll_area.bottom, self.attr)
         if not self.mode.is_text_mode:
@@ -511,7 +532,9 @@ class TextScreen(object):
             if not self.text.pages[self.apagenum].row[self.current_row-1].wrap:
                 return
             # else: LF case; scroll up without changing attributes
-            self.text.pages[self.apagenum].row[self.current_row-1].wrap = self.text.pages[self.apagenum].row[self.current_row].wrap
+            self.text.pages[self.apagenum].row[self.current_row-1].wrap = (
+                self.text.pages[self.apagenum].row[self.current_row].wrap
+            )
             self.scroll(self.current_row + 1)
             # redraw from the LF
             self._rewrite_for_delete(text[1:] + b' ' * cwidth)
@@ -519,7 +542,10 @@ class TextScreen(object):
             # rewrite the contents (with the current attribute!)
             self._rewrite_for_delete(text[cwidth:] + b' ' * cwidth)
             # if last row was empty, scroll up.
-            if self.text.pages[self.apagenum].row[lastrow-1].end == 0 and self.text.pages[self.apagenum].row[lastrow-2].wrap:
+            if (
+                    self.text.pages[self.apagenum].row[lastrow-1].end == 0 and
+                    self.text.pages[self.apagenum].row[lastrow-2].wrap
+                ):
                 self.text.pages[self.apagenum].row[lastrow-2].wrap = False
                 self.scroll(lastrow)
         # restore original position
@@ -527,9 +553,11 @@ class TextScreen(object):
 
     def _rewrite_for_delete(self, text):
         """Rewrite text contents (with the current attribute)."""
-        for c in text:
+        for c in iterchar(text):
             if c == b'\n':
-                self.put_char_attr(self.apagenum, self.current_row, self.current_col, b' ', self.attr)
+                self.put_char_attr(
+                    self.apagenum, self.current_row, self.current_col, b' ', self.attr
+                )
                 self.text.pages[self.apagenum].row[self.current_row-1].end = self.current_col-1
                 break
             else:
@@ -544,50 +572,84 @@ class TextScreen(object):
                 # adjust row end
                 self.text.pages[self.apagenum].row[self.current_row-1].end = self.current_col - 2
 
-    def insert_fullchars(self, row, col, sequence, attr):
-        """Insert one or more single- or double-width characters at the current position."""
-        start_row, start_col = row, col
-        for c in sequence:
-            while True:
-                therow = self.text.pages[self.apagenum].row[row-1]
-                therow.buf.insert(col-1, (c, attr))
-                if therow.end < self.mode.width:
-                    therow.buf.pop()
-                    if therow.end > col-1:
-                        therow.end += 1
-                    else:
-                        therow.end = col
-                    break
+    def insert_fullchars(self, sequence):
+        """Insert one or more single- or double-width characters and adjust cursor."""
+        # insert one at a time at cursor location
+        # to let cursor position logic deal with scrolling
+        for c in iterchar(sequence):
+            if self._insert_fullchar_at(self.current_row, self.current_col, c, self.attr):
+                # move cursor by one character
+                # this will move to next row when necessary
+                self.set_pos(self.current_row, self.current_col+1)
+
+    def _insert_fullchar_at(self, row, col, c, attr):
+        """Insert one single- or double-width character at the given position."""
+        therow = self.text.pages[self.apagenum].row[row-1]
+        if therow.end < self.mode.width:
+            therow.buf.insert(col-1, (c, attr))
+            c, attr = therow.buf.pop()
+            if therow.end > col-1:
+                therow.end += 1
+            else:
+                therow.end = col
+            self._redraw_row(col-1, row)
+            if  therow.wrap and therow.end == self.mode.width:
+                self.scroll_down(row+1)
+                therow.wrap = True
+            return True
+        else:
+            # pushing the end of the row past the screen edge
+            # if we're not a wrapping line, make space by scrolling
+            if not therow.wrap and row < self.scroll_area.bottom:
+                self.scroll_down(row+1)
+                therow.wrap = True
+            if row >= self.scroll_area.bottom:
+                # once the end of the line hits the bottom, start scrolling the start of the line up
+                # until that hist the top of the screen. After that, stop inserting & drop chars
+                start = self.text.find_start_of_line(self.apagenum, self.current_row)
+                if start > self.scroll_area.top:
+                    self.scroll()
+                    row -= 1
                 else:
-                    if row == self.scroll_area.bottom:
-                        self.scroll()
-                        row -= 1
-                    if not therow.wrap and row < self.mode.height:
-                        self.scroll_down(row+1)
-                        therow.wrap = True
-                    c, attr = therow.buf.pop()
-                    row += 1
-                    col = 1
-            col += 1
-        self._redraw_row(start_col-1, start_row)
+                    return False
+            therow.buf.insert(col-1, (c, attr))
+            c, attr = therow.buf.pop()
+            self._redraw_row(col-1, row)
+            # insert the character in the next row
+            return self._insert_fullchar_at(row+1, 1, c, attr)
 
     def line_feed(self):
         """Move the remainder of the line to the next row and wrap (LF)."""
-        row, col = self.current_row, self.current_col
-        if col < self.text.pages[self.apagenum].row[row-1].end:
-            self.insert_fullchars(row, col, b' ' * (self.mode.width-col+1), self.attr)
-            self.text.pages[self.apagenum].row[row-1].end = col - 1
+        if self.current_col < self.text.pages[self.apagenum].row[self.current_row-1].end:
+            # insert characters, preserving cursor position
+            cursor = self.current_row, self.current_col
+            self.insert_fullchars(b' ' * (self.mode.width-self.current_col+1))
+            self.set_pos(*cursor, scroll_ok=False)
+            # adjust end of line and wrapping flag - LF connects lines like word wrap
+            self.text.pages[self.apagenum].row[self.current_row-1].end = self.current_col - 1
+            self.text.pages[self.apagenum].row[self.current_row-1].wrap = True
+            # cursor stays in place after line feed!
         else:
-            while (self.text.pages[self.apagenum].row[row-1].wrap and row < self.scroll_area.bottom):
-                row += 1
-            if row >= self.scroll_area.bottom:
-                self.scroll()
+            # find last row in logical line
+            end = self.text.find_end_of_line(self.apagenum, self.current_row)
+            # if the logical line hits the bottom, start scrolling up to make space...
+            if end >= self.scroll_area.bottom:
+                # ... until the it also hits the top; then do nothing
+                start = self.text.find_start_of_line(self.apagenum, self.current_row)
+                if start > self.scroll_area.top:
+                    self.scroll()
+                else:
+                    return
             # self.current_row has changed, don't use row var
             if self.current_row < self.mode.height:
                 self.scroll_down(self.current_row+1)
-        # LF connects lines like word wrap
-        self.text.pages[self.apagenum].row[self.current_row-1].wrap = True
-        self.set_pos(self.current_row+1, 1)
+                # if we were already a wrapping row, make sure the new empty row wraps
+                #if self.text.pages[self.apagenum].row[self.current_row-1].wrap:
+                #    self.text.pages[self.apagenum].row[self.current_row].wrap = True
+            # ensure the current row now wraps
+            self.text.pages[self.apagenum].row[self.current_row-1].wrap = True
+            # cursor moves to start of next line
+            self.set_pos(self.current_row+1, 1)
 
     ###########################################################################
     # vpage text retrieval
@@ -602,10 +664,12 @@ class TextScreen(object):
     def copy_clipboard(self, start_row, start_col, stop_row, stop_col, is_mouse_selection):
         """Copy selected screen area to clipboard."""
         clips = self.text.get_text_logical(
-                self.vpagenum, start_row, start_col, stop_row, stop_col)
+            self.vpagenum, start_row, start_col, stop_row, stop_col
+        )
         text = u'\n'.join(self.codepage.str_to_unicode(clip) for clip in clips)
         self.queues.video.put(signals.Event(
-                signals.VIDEO_SET_CLIPBOARD_TEXT, (text, is_mouse_selection)))
+                signals.VIDEO_SET_CLIPBOARD_TEXT, (text, is_mouse_selection)
+        ))
 
     ###########################################################################
     # text screen callbacks
@@ -695,7 +759,10 @@ class TextScreen(object):
         if start is None and stop is None:
             self.scroll_area.unset()
         else:
-            max_line = 25 if (self.capabilities in ('pcjr', 'tandy') and not self.bottom_bar.visible) else 24
+            if self.capabilities in ('pcjr', 'tandy') and not self.bottom_bar.visible:
+                max_line = 25
+            else:
+                max_line = 24
             error.range_check(1, max_line, start, stop)
             error.throw_if(stop < start)
             self._set_scroll_area(start, stop)
