@@ -2,7 +2,7 @@
 PC-BASIC - dos.py
 Operating system shell and environment
 
-(c) 2013--2018 Rob Hagemans
+(c) 2013--2019 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
@@ -15,7 +15,7 @@ from collections import deque
 import subprocess
 from subprocess import Popen, PIPE
 
-from ..compat import SHELL_ENCODING, HIDE_WINDOW, split_quoted
+from ..compat import SHELL_ENCODING, HIDE_WINDOW, split_quoted, getenvu, setenvu, iterenvu
 from .codepage import CONTROL
 from .base import error
 from . import values
@@ -29,45 +29,65 @@ from . import values
 SHELL_COMMAND_SWITCH = u'/C'
 
 
-
-#########################################
+##########################################
 # calling shell environment
 
 class Environment(object):
     """Handle environment changes."""
 
-    def __init__(self, values):
+    def __init__(self, values, codepage):
         """Initialise."""
         self._values = values
+        self._codepage = codepage
+
+    def _setenv(self, key, value):
+        """Set environment (bytes) key to (bytes) value."""
+        assert isinstance(key, bytes), type(key)
+        assert isinstance(value, bytes), type(value)
+        ukey = self._codepage.str_to_unicode(key)
+        uvalue = self._codepage.str_to_unicode(value)
+        setenvu(ukey, uvalue)
+
+    def _getenv(self, key):
+        """Get environment (bytes) value or b''."""
+        assert isinstance(key, bytes), type(key)
+        ukey = self._codepage.str_to_unicode(key)
+        return self._codepage.from_unicode(getenvu(ukey, u''))
+
+    def _getenv_item(self, index):
+        """Get environment (bytes) 'key=value' or b'', by zero-based index."""
+        envlist = list(iterenvu())
+        try:
+            ukey = envlist[index]
+        except IndexError:
+            return b''
+        else:
+            key = self._codepage.from_unicode(ukey)
+            value = self._codepage.from_unicode(getenvu(ukey, u''))
+            return b'%s=%s' % (key, value)
 
     def environ_(self, args):
         """ENVIRON$: get environment string."""
         expr, = args
         if isinstance(expr, values.String):
-            parm = expr.to_str()
-            if not parm:
+            key = expr.to_str()
+            if not key:
                 raise error.BASICError(error.IFC)
-            envstr = os.getenv(parm) or b''
+            result = self._getenv(key)
         else:
-            expr = values.to_int(expr)
-            error.range_check(1, 255, expr)
-            envlist = list(os.environ)
-            if expr > len(envlist):
-                envstr = ''
-            else:
-                envstr = '%s=%s' % (envlist[expr-1], os.getenv(envlist[expr-1]))
-        return self._values.new_string().from_str(envstr)
+            index = values.to_int(expr)
+            error.range_check(1, 255, index)
+            result = self._getenv_item(index-1)
+        return self._values.new_string().from_str(result)
 
     def environ_statement_(self, args):
         """ENVIRON: set environment string."""
         envstr = values.next_string(args)
         list(args)
-        eqs = envstr.find('=')
+        eqs = envstr.find(b'=')
         if eqs <= 0:
             raise error.BASICError(error.IFC)
-        envvar = str(envstr[:eqs])
-        val = str(envstr[eqs+1:])
-        os.environ[envvar] = val
+        self._setenv(envstr[:eqs], envstr[eqs+1:])
 
 
 #########################################
@@ -103,7 +123,7 @@ class Shell(object):
         """Run a SHELL subprocess."""
         logging.debug('Executing SHELL command %r', command)
         if not self._shell:
-            logging.warning(b'SHELL statement not enabled: no command interpreter specified.')
+            logging.warning('SHELL statement not enabled: no command interpreter specified.')
             raise error.BASICError(error.IFC)
         cmd = split_quoted(self._shell)
         if command:
@@ -112,8 +132,9 @@ class Shell(object):
         work_dir = self._files.get_native_cwd()
         try:
             p = Popen(
-                    cmd, shell=False, cwd=work_dir,
-                    stdin=PIPE, stdout=PIPE, stderr=PIPE, startupinfo=HIDE_WINDOW)
+                cmd, shell=False, cwd=work_dir,
+                stdin=PIPE, stdout=PIPE, stderr=PIPE, startupinfo=HIDE_WINDOW
+            )
         except (EnvironmentError, UnicodeEncodeError) as e:
             logging.warning(u'SHELL: command interpreter `%s` not accessible: %s', self._shell, e)
             raise error.BASICError(error.IFC)
@@ -164,8 +185,8 @@ class Shell(object):
                 continue
             elif c in (b'\r', b'\n'):
                 # put sentinel on queue
-                shell_output.append('')
-                shell_cerr.append('')
+                shell_output.append(b'')
+                shell_cerr.append(b'')
                 # send the command
                 self._send_input(p.stdin, word)
                 word = []
