@@ -2,13 +2,15 @@
 PC-BASIC - machine.py
 Machine emulation and memory model
 
-(c) 2013--2018 Rob Hagemans
+(c) 2013--2019 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
 import struct
 import logging
 
+from ..compat import iteritems, int2byte
+
 from ..metadata import NAME, VERSION, COPYRIGHT
 from .base import error
 from . import values
@@ -16,8 +18,9 @@ from . import devices
 
 
 # ROM copyright notice
-NOTICE = bytearray(b'%s %s\r%s\r' %
-            tuple(_.encode('ascii', 'ignore') for _ in (NAME, VERSION, COPYRIGHT)))
+NOTICE = bytearray(
+    b'%s %s\r%s\r' % tuple(_.encode('ascii', 'ignore') for _ in (NAME, VERSION, COPYRIGHT))
+)
 
 
 ###############################################################################
@@ -38,10 +41,11 @@ class MachinePorts(object):
         self._display = display
         # parallel port base address:
         # http://retired.beyondlogic.org/spp/parallel.htm
-        # 3BCh - 3BFh  Used for Parallel Ports which were incorporated on to Video Cards - Doesn't support ECP addresses
+        # 3BCh - 3BFh  Used for Parallel Ports which were incorporated on to Video Cards
+        #              - Doesn't support ECP addresses
         # 378h - 37Fh  Usual Address For LPT 1
         # 278h - 27Fh  Usual Address For LPT 2
-        self.lpt_device = [files.get_device('LPT1:'), files.get_device('LPT2:')]
+        self.lpt_device = [files.get_device(b'LPT1:'), files.get_device(b'LPT2:')]
         # serial port base address:
         # http://www.petesqbsite.com/sections/tutorials/zines/qbnews/9-com_ports.txt
         #            COM1             &H3F8
@@ -49,7 +53,7 @@ class MachinePorts(object):
         #            COM3             &H3E8 (not implemented)
         #            COM4             &H2E8 (not implemented)
         self.com_base = {0x3f8: 0, 0x2f8: 1}
-        self.com_device = [files.get_device('COM1:'), files.get_device('COM2:')]
+        self.com_device = [files.get_device(b'COM1:'), files.get_device(b'COM2:')]
         self.com_enable_baud_write = [False, False]
         self.com_baud_divisor = [0, 0]
         self.com_break = [False, False]
@@ -109,7 +113,7 @@ class MachinePorts(object):
             # serial port machine ports
             # http://www.qb64.net/wiki/index.php/Port_Access_Libraries#Serial_Communication_Registers
             # http://control.com/thread/1026221083
-            for base_addr, com_port_nr in self.com_base.iteritems():
+            for base_addr, com_port_nr in iteritems(self.com_base):
                 com_port = self.com_device[com_port_nr]
                 if not com_port.available():
                     continue
@@ -118,7 +122,7 @@ class MachinePorts(object):
                     _, parity, bytesize, stopbits = com_port.get_params()
                     value = self.com_enable_baud_write[com_port_nr] * 0x80
                     value += self.com_break[com_port_nr] * 0x40
-                    value += {'S': 0x38, 'M': 0x28, 'E': 0x18, 'O': 0x8, 'N': 0}[parity]
+                    value += {b'S': 0x38, b'M': 0x28, b'E': 0x18, b'O': 0x8, b'N': 0}[parity]
                     if stopbits > 1:
                         value += 0x4
                     value += bytesize - 5
@@ -172,16 +176,17 @@ class MachinePorts(object):
                 return
             if addr - base_addr[lpt_port_nr] == 0:
                 # set data port
-                self.lpt_device[lpt_port_nr].stream.write(chr(val))
+                self.lpt_device[lpt_port_nr].stream.write(int2byte(val))
             else:
                 # set control port
                 self.lpt_device[lpt_port_nr].stream.set_control(
-                    select=val & 0x8, init=val&0x4, lf=val&0x2, strobe=val&0x1)
+                    select=val & 0x8, init=val&0x4, lf=val&0x2, strobe=val&0x1
+                )
         else:
             # serial port machine ports
             # http://www.qb64.net/wiki/index.php/Port_Access_Libraries#Serial_Communication_Registers
             # http://control.com/thread/1026221083
-            for base_addr, com_port_nr in self.com_base.iteritems():
+            for base_addr, com_port_nr in iteritems(self.com_base):
                 com_port = self.com_device[com_port_nr]
                 if not com_port.available():
                     continue
@@ -189,9 +194,13 @@ class MachinePorts(object):
                 # (among other things that aren't implemented)
                 if addr in (base_addr, base_addr+1) and self.com_enable_baud_write[com_port_nr]:
                     if addr == base_addr:
-                        self.com_baud_divisor[com_port_nr] = (self.com_baud_divisor[com_port_nr] & 0xff00) + val
+                        self.com_baud_divisor[com_port_nr] = (
+                            (self.com_baud_divisor[com_port_nr] & 0xff00) + val
+                        )
                     elif addr == base_addr + 1:
-                        self.com_baud_divisor[com_port_nr] = val*0x100 + (self.com_baud_divisor[com_port_nr] & 0xff)
+                        self.com_baud_divisor[com_port_nr] = (
+                            val*0x100 + (self.com_baud_divisor[com_port_nr] & 0xff)
+                        )
                     if self.com_baud_divisor[com_port_nr]:
                         baudrate, parity, bytesize, stopbits = com_port.get_params()
                         baudrate = 115200 // self.com_baud_divisor[com_port_nr]
@@ -204,7 +213,7 @@ class MachinePorts(object):
                     # break condition
                     self.com_break[com_port_nr] = (val & 0x40) != 0
                     # parity
-                    parity = {0x38:'S', 0x28:'M', 0x18:'E', 0x8:'O', 0:'N'}[val&0x38]
+                    parity = {0x38: b'S', 0x28: b'M', 0x18: b'E', 0x8: b'O', 0: b'N'}[val&0x38]
                     # stopbits
                     if val & 0x4:
                         # 2 or 1.5 stop bits
@@ -314,7 +323,7 @@ class Memory(object):
         if offset is not None:
             offset = values.to_int(offset, unsigned=True)
         list(args)
-        with self._files.open(0, name, filetype='M', mode='I') as g:
+        with self._files.open(0, name, filetype=b'M', mode=b'I') as g:
             # size gets ignored; even the \x1a at the end gets dumped onto the screen.
             seg = g.seg
             if offset is None:
@@ -337,14 +346,17 @@ class Memory(object):
         offset = values.to_int(next(args), unsigned=True)
         length = values.to_int(next(args), unsigned=True)
         list(args)
-        with self._files.open(0, name, filetype='M', mode='O',
-                    seg=self.segment, offset=offset, length=length) as g:
+        with self._files.open(
+                    0, name, filetype=b'M', mode=b'O',
+                    seg=self.segment, offset=offset, length=length
+                ) as g:
             addr = self.segment * 0x10 + offset
-            g.write(str(self._get_memory_block(addr, length)))
+            g.write(bytes(self._get_memory_block(addr, length)))
             # Tandys repeat the header at the end of the file
             if self._syntax == 'tandy':
-                g.write(devices.TYPE_TO_MAGIC['M'] +
-                        struct.pack('<HHH', self.segment, offset, length))
+                g.write(
+                    devices.TYPE_TO_MAGIC[b'M'] + struct.pack('<HHH', self.segment, offset, length)
+                )
 
     def def_seg_(self, args):
         """DEF SEG: Set segment."""
@@ -367,7 +379,7 @@ class Memory(object):
     def call_(self, args):
         """CALL or CALLS: Call machine language procedure."""
         addr_var = next(args)
-        if self._memory.complete_name(addr_var)[-1] == values.STR:
+        if self._memory.complete_name(addr_var)[-1:] == values.STR:
             # type mismatch
             raise error.BASICError(error.TYPE_MISMATCH)
         list(args)
@@ -424,7 +436,7 @@ class Memory(object):
             addr += video_len
             length -= video_len
         for a in range(addr, addr+length):
-            block += chr(max(0, self._get_memory(a)))
+            block.append(max(0, self._get_memory(a)))
         return block
 
     def _set_memory_block(self, addr, buf):
@@ -545,12 +557,14 @@ class Memory(object):
         #   "(PEEK (1041) AND 192)/64" WILL PROVIDE NUMBER OF PRINTERS INSTALLED.
         elif addr == 1041:
             return (
-                2 * (self._files.device_available('COM1:') +
-                    self._files.device_available('COM2:')) +
-                16 +
-                64 * (self._files.device_available('LPT1:') +
-                    self._files.device_available('LPT2:') +
-                    self._files.device_available('LPT3:'))
+                2 * (
+                    self._files.device_available(b'COM1:') +
+                    self._files.device_available(b'COM2:')
+                ) + 16 + 64 * (
+                    self._files.device_available(b'LPT1:') +
+                    self._files.device_available(b'LPT2:') +
+                    self._files.device_available(b'LPT3:')
+                )
             )
         # &h40:&h17 keyboard flag
         # &H80 - Insert state active
@@ -585,12 +599,12 @@ class Memory(object):
         elif addr == 1053:
             return (self.keyboard.buf.stop*2 + self.key_buffer_offset) // 256
         elif addr in range(1024+self.key_buffer_offset, 1024+self.key_buffer_offset+32):
-            index = (addr-1024-self.key_buffer_offset)//2
-            odd = (addr-1024-self.key_buffer_offset)%2
+            index = (addr-1024-self.key_buffer_offset) // 2
+            odd = (addr-1024-self.key_buffer_offset) % 2
             c, scan = self.keyboard.buf.ring_read(index)
             if odd:
                 return scan
-            elif c == '':
+            elif c == b'':
                 return 0
             else:
                 # however, arrow keys (all extended scancodes?) give 0xe0 instead of 0
@@ -607,11 +621,13 @@ class Memory(object):
             elif self.display.mode.name == '320x200x4':
                 return 4 + cval
             else:
-                mode_num = {'640x200x2': 6, '160x200x16': 8, '320x200x16pcjr': 9,
+                mode_num = {
+                    '640x200x2': 6, '160x200x16': 8, '320x200x16pcjr': 9,
                     '640x200x4': 10, '320x200x16': 13, '640x200x16': 14,
                     '640x350x4': 15, '640x350x16': 16, '640x400x2': 0x40,
-                    '320x200x4pcjr': 4 }
+                    '320x200x4pcjr': 4
                     # '720x348x2': ? # hercules - unknown
+                }
                 try:
                     return mode_num[self.display.mode.name]
                 except KeyError:
@@ -646,16 +662,20 @@ class Memory(object):
         elif addr == 1125:
             # bit 0: only in text mode?
             # bit 2: should this be colorswitch or colorburst_is_enabled?
-            return ((self.display.mode.width == 80) * 1 +
-                    (not self.display.mode.is_text_mode) * 2 +
-                     self.display.colorswitch * 4 + 8 +
-                     (self.display.mode.name == '640x200x2') * 16 +
-                     self.blink_enabled * 32)
+            return (
+                (self.display.mode.width == 80) * 1 +
+                (not self.display.mode.is_text_mode) * 2 +
+                self.display.colorswitch * 4 + 8 +
+                (self.display.mode.name == '640x200x2') * 16 +
+                self.blink_enabled * 32
+            )
         # 1126 color
         elif addr == 1126:
             if self.display.mode.name == '320x200x4':
-                return (self.display.palette.get_entry(0)
-                        + 32 * self.display.video.cga4_palette_num)
+                return (
+                    self.display.palette.get_entry(0)
+                    + 32 * self.display.video.cga4_palette_num
+                )
             elif self.display.mode.is_text_mode:
                 return self.display.border_attr % 16
                 # not implemented: + 16 "if current color specified through
@@ -673,21 +693,23 @@ class Memory(object):
         elif addr == 1050:
             # keyboard ring buffer starts at n+1024; lowest 1054
             self.keyboard.buf.ring_set_boundaries(
-                    (value - self.key_buffer_offset) // 2,
-                    self.keyboard.buf.stop)
+                (value - self.key_buffer_offset) // 2,
+                self.keyboard.buf.stop
+            )
         elif addr == 1052:
             # ring buffer ends at n + 1023
             self.keyboard.buf.ring_set_boundaries(
-                    self.keyboard.buf.start,
-                    (value - self.key_buffer_offset) // 2)
+                self.keyboard.buf.start,
+                (value - self.key_buffer_offset) // 2
+            )
         elif addr in range(1024+self.key_buffer_offset, 1024+self.key_buffer_offset+32):
-            index = (addr-1024-self.key_buffer_offset)//2
-            odd = (addr-1024-self.key_buffer_offset)%2
+            index = (addr-1024-self.key_buffer_offset) // 2
+            odd = (addr-1024-self.key_buffer_offset) % 2
             c, scan = self.keyboard.buf.ring_read(index)
             if odd:
                 scan = value
             elif value in (0, 0xe0):
-                c = ''
+                c = b''
             else:
-                c = chr(value)
+                c = int2byte(value)
             self.keyboard.buf.ring_write(index, c, scan)
