@@ -2,7 +2,7 @@
 PC-BASIC - numbers.py
 Integer and Floating Point values
 
-(c) 2013--2018 Rob Hagemans
+(c) 2013--2019 Rob Hagemans
 This file is released under the GNU GPL version 3.
 """
 
@@ -22,9 +22,12 @@ This file is released under the GNU GPL version 3.
 # The exponent is biased by 128.
 # There is an assumed 1 bit after the radix point (so the assumed mantissa is 0.1ffff... where f's are the fraction bits)
 
+import binascii
 import struct
 import math
 
+from ...compat import iterchar, int2byte
+
 from ..base import tokens as tk
 from ..base import error
 
@@ -53,15 +56,15 @@ class Value(object):
         self._buffer = memoryview(buffer)
         self._values = values
 
-    def __str__(self):
+    def __repr__(self):
         """String representation for debugging."""
         try:
-            return b'%s[%s %s]' % (self.sigil, bytes(self.to_bytes()).encode('hex'), repr(self.to_value()))
+            return '%s[%s %r]' % (
+                self.sigil, binascii.hexlify(self.to_bytes()),
+                self.to_value()
+            )
         except Exception:
-            return b'%s[%s <detached>]' % (self.sigil, bytes(self.to_bytes()).encode('hex'))
-
-    __repr__ = __str__
-
+            return '%s[%s <detached>]' % (self.sigil, binascii.hexlify(self.to_bytes()))
 
     def __getstate__(self):
         # can't pickle memoryview
@@ -153,11 +156,16 @@ class Integer(Number):
 
     def is_negative(self):
         """Value is negative."""
-        return (ord(self._buffer[-1]) & 0x80) != 0
+        return (bytearray(self._buffer)[-1] & 0x80) != 0
 
     def sign(self):
         """Sign of value."""
-        return -1 if (ord(self._buffer[-1]) & 0x80) != 0 else (0 if self._buffer == b'\0\0' else 1)
+        if (bytearray(self._buffer)[-1] & 0x80) != 0:
+            return -1
+        elif bytearray(self._buffer) == bytearray(b'\0\0'):
+            return 0
+        else:
+            return 1
 
     def to_int(self, unsigned=False):
         """Return value as Python int."""
@@ -203,13 +211,13 @@ class Integer(Number):
 
     def to_token(self):
         """Return signed value as integer token."""
-        if self._buffer[1] == b'\0':
-            byte = ord(self._buffer[0])
+        if bytearray(self._buffer)[1] == 0:
+            byte = bytearray(self._buffer)[0]
             # although there is a one-byte token for '10', we don't write it.
             if byte < 10:
-                return chr(ord(tk.C_0) + byte)
+                return int2byte(ord(tk.C_0) + byte)
             else:
-                return tk.T_BYTE + self._buffer[0]
+                return tk.T_BYTE + int2byte(bytearray(self._buffer)[0])
         else:
             return tk.T_INT + self._buffer.tobytes()
 
@@ -227,13 +235,13 @@ class Integer(Number):
 
     def from_token(self, token):
         """Set value to signed or unsigned integer token."""
-        d = bytes(token)[0]
-        if d in (tk.T_OCT, tk.T_HEX, tk.T_INT, tk.T_UINT):
+        d = bytearray(token)[0]
+        if d in (ord(_c) for _c in (tk.T_OCT, tk.T_HEX, tk.T_INT, tk.T_UINT)):
             self._buffer[:] = token[-2:]
-        elif d == tk.T_BYTE:
-            self._buffer[:] = token[-1] + b'\0'
-        elif tk.C_0 <= d <= tk.C_10:
-            self._buffer[:] = chr(ord(d) - 0x11) + b'\0'
+        elif d == ord(tk.T_BYTE):
+            self._buffer[:] = token[-1:] + b'\0'
+        elif ord(tk.C_0) <= d <= ord(tk.C_10):
+            self._buffer[:] = int2byte(d - 0x11) + b'\0'
         else:
             raise ValueError('%s is not an Integer token.' % repr(token))
         return self
@@ -244,16 +252,16 @@ class Integer(Number):
         """Convert integer to str in octal representation."""
         if self.is_zero():
             return b'0'
-        return oct(self.to_int(unsigned=True))[1:]
+        return b'%o' % (self.to_int(unsigned=True),)
 
     def to_hex(self):
         """Convert integer to str in hex representation."""
-        return hex(self.to_int(unsigned=True))[2:].upper()
+        return b'%X' % (self.to_int(unsigned=True),)
 
     def to_str(self, leading_space, type_sign):
         """Convert integer to str in decimal representation."""
-        intstr = bytes(self.to_int())
-        if leading_space and intstr[0] != b'-':
+        intstr = b'%d' % self.to_int()
+        if leading_space and intstr[0:1] != b'-':
             return b' ' + intstr
         else:
             return intstr
@@ -261,7 +269,7 @@ class Integer(Number):
     def from_oct(self, oct_repr):
         """Convert str in octal representation to integer."""
         # oct representations may be interrupted by blanks
-        val = int(oct_repr.strip(BLANKS), 8) if oct_repr else 0
+        val = int(oct_repr.strip(BLANKS), 8) if oct_repr.strip(BLANKS) else 0
         return self.from_int(val, unsigned=True)
 
     def from_hex(self, hex_repr):
@@ -272,7 +280,13 @@ class Integer(Number):
 
     def from_str(self, dec_repr):
         """Convert str in decimal representation to integer."""
-        return self.from_int(int(dec_repr.strip(BLANKS)))
+        valstr = dec_repr.strip(BLANKS)
+        if set(valstr) - set(tk.DIGITS):
+            raise ValueError(
+                'string %r contains non-digits, cannot convert to Integer.' %
+                (valstr,)
+            )
+        return self.from_int(int(valstr))
 
     # operations
 
@@ -290,35 +304,39 @@ class Integer(Number):
 
     def ineg(self):
         """Negate in-place."""
-        if self._buffer == '\x00\x80':
+        if self._buffer == b'\x00\x80':
             raise error.BASICError(error.OVERFLOW)
-        lsb = (ord(self._buffer[0]) ^ 0xff) + 1
-        msb = ord(self._buffer[1]) ^ 0xff
+        lsb = (bytearray(self._buffer)[0] ^ 0xff) + 1
+        msb = bytearray(self._buffer)[1] ^ 0xff
         # apply carry
         if lsb > 0xff:
             lsb -= 0x100
             msb += 1
         # ignore overflow, since -0 == 0
-        self._buffer[:] = chr(lsb) + chr(msb & 0xff)
+        self._buffer[:] = bytearray([lsb, msb & 0xff])
         return self
 
     def iabs(self):
         """Absolute value in-place."""
-        if (ord(self._buffer[-1]) & 0x80):
+        if bytearray(self._buffer)[-1] & 0x80:
             return self.ineg()
 
     def iadd(self, rhs):
         """Add another Integer in-place."""
-        lsb = ord(self._buffer[0]) + ord(rhs._buffer[0])
-        msb = ord(self._buffer[1]) + ord(rhs._buffer[1])
+        lsb = bytearray(self._buffer)[0] + bytearray(rhs._buffer)[0]
+        msb = bytearray(self._buffer)[1] + bytearray(rhs._buffer)[1]
         # apply carry
         if lsb > 0xff:
             lsb -= 0x100
             msb += 1
         # overflow if signs were equal and have changed
-        if (ord(self._buffer[1]) > 0x7f) == (ord(rhs._buffer[1]) > 0x7f) != (msb > 0x7f):
+        if (
+                (bytearray(self._buffer)[1] > 0x7f) ==
+                (bytearray(rhs._buffer)[1] > 0x7f) !=
+                (msb > 0x7f)
+            ):
             raise error.BASICError(error.OVERFLOW)
-        self._buffer[:] = chr(lsb) + chr(msb & 0xff)
+        self._buffer[:] = bytearray([lsb, msb & 0xff])
         return self
 
     def isub(self, rhs):
@@ -371,17 +389,17 @@ class Integer(Number):
         if isinstance(rhs, Float):
             # upgrade to Float
             return rhs.new().from_integer(self).gt(rhs)
-        isneg = ord(self._buffer[-1]) & 0x80
-        if isneg != (ord(rhs._buffer[-1]) & 0x80):
+        isneg = bytearray(self._buffer)[-1] & 0x80
+        if isneg != (bytearray(rhs._buffer)[-1] & 0x80):
             return not(isneg)
         # compute the unsigned (not absolute!) >
-        lmsb = ord(self._buffer[1]) & 0x7f
-        rmsb = ord(rhs._buffer[1]) & 0x7f
+        lmsb = bytearray(self._buffer)[1] & 0x7f
+        rmsb = bytearray(rhs._buffer)[1] & 0x7f
         if lmsb > rmsb:
             return True
         elif lmsb < rmsb:
             return False
-        return ord(self._buffer[0]) > ord(rhs._buffer[0])
+        return bytearray(self._buffer)[0] > bytearray(rhs._buffer)[0]
 
     def eq(self, rhs):
         """Equals."""
@@ -407,15 +425,19 @@ class Float(Number):
 
     def is_zero(self):
         """Value is zero."""
-        return self._buffer[-1] == b'\0'
+        return bytearray(self._buffer)[-1] == 0
 
     def is_negative(self):
         """Value is negative."""
-        return self._buffer[-2] >= b'\x80'
+        return bytearray(self._buffer)[-2] >= 0x80
 
     def sign(self):
         """Sign of value."""
-        return 0 if self._buffer[-1] == b'\0' else (-1 if (ord(self._buffer[-2]) & 0x80) != 0 else 1)
+        if bytearray(self._buffer)[-1] == 0:
+            return 0
+        elif (bytearray(self._buffer)[-2] & 0x80) != 0:
+            return -1
+        return 1
 
     # BASIC type conversions
 
@@ -431,11 +453,11 @@ class Float(Number):
 
     def to_value(self):
         """Return value as Python float."""
-        exp = ord(self._buffer[-1]) - self._bias
-        if exp == -self._bias:
+        exp = bytearray(self._buffer)[-1] - self._bias
+        if exp == -self._bias:  # pylint: disable=invalid-unary-operand-type
             return 0.
         # unpack as unsigned long int
-        man = struct.unpack(self._intformat, bytearray(self._buffer[:-1]) + b'\0')[0]
+        man = struct.unpack(self._intformat, bytearray(self._buffer)[:-1] + b'\0')[0]
         # prepend assumed bit and apply sign
         if man & self._signmask:
             man = -man
@@ -456,8 +478,10 @@ class Float(Number):
         man, exp = self._bring_to_range(man, exp, self._posmask, self._mask)
         if not self._check_limits(exp, neg):
             return self
-        struct.pack_into(self._intformat, self._buffer, 0, man & (self._mask if neg else self._posmask))
-        self._buffer[-1] = chr(exp)
+        struct.pack_into(
+            self._intformat, self._buffer, 0, man & (self._mask if neg else self._posmask)
+        )
+        self._buffer[-1:] = int2byte(exp)
         return self
 
 
@@ -482,8 +506,10 @@ class Float(Number):
             man, exp = self._bring_to_range(abs(in_int), self._bias, self._posmask, self._mask)
             if not self._check_limits(exp, neg):
                 return self
-            struct.pack_into(self._intformat, self._buffer, 0, man & (self._mask if neg else self._posmask))
-            self._buffer[-1] = chr(exp)
+            struct.pack_into(
+                self._intformat, self._buffer, 0, man & (self._mask if neg else self._posmask)
+            )
+            self._buffer[-1:] = int2byte(exp)
         return self
 
     def to_int_truncate(self):
@@ -501,12 +527,12 @@ class Float(Number):
 
     def ineg(self):
         """Negate in-place."""
-        self._buffer[-2] = chr(ord(self._buffer[-2]) ^ 0x80)
+        self._buffer[-2:-1] = int2byte(bytearray(self._buffer)[-2] ^ 0x80)
         return self
 
     def iabs(self):
         """Absolute value in-place."""
-        self._buffer[-2] = chr(ord(self._buffer[-2]) & 0x7F)
+        self._buffer[-2:-1] = int2byte(bytearray(self._buffer)[-2] & 0x7F)
         return self
 
     def iround(self):
@@ -615,25 +641,27 @@ class Float(Number):
     def to_str(self, leading_space, type_sign):
         """Convert to string representation."""
         if self.is_zero():
-            return (' ' * leading_space) + '0' + (self.sigil * type_sign)
+            return (b' ' * leading_space) + b'0' + (self.sigil * type_sign)
         # sign or leading space
-        sign = ''
+        sign = b''
         if self.is_negative():
-            sign = '-'
+            sign = b'-'
         elif leading_space:
-            sign = ' '
+            sign = b' '
         mantissa, exp10 = self.to_decimal()
         digitstr = _get_digits(mantissa, self.digits, remove_trailing=True)
         # exponent for scientific notation
         exp10 += self.digits - 1
         if exp10 > self.digits - 1 or len(digitstr)-exp10 > self.digits + 1:
             # use scientific notation
-            # this is self.to_str_scientific(n_before=1, n_decimals=self.digits-1, force_dot=False, group_digits=False)
+            # this is self.to_str_scientific(
+            #     n_before=1, n_decimals=self.digits-1, force_dot=False, group_digits=False)
             valstr = self._scientific_notation(digitstr, exp10, digits_to_dot=1, force_dot=False)
         else:
             # use decimal notation
             # usual: n_decimals = -exp10
-            # this is self.to_str_fixed(self, n_decimals=-exp10, force_dot=False, group_digits=False) but with type_sign
+            # this is self.to_str_fixed
+            #     (self, n_decimals=-exp10, force_dot=False, group_digits=False) but with type_sign
             valstr = self._decimal_notation(digitstr, exp10, type_sign, force_dot=False)
         return sign + valstr
 
@@ -690,30 +718,37 @@ class Float(Number):
 
     def to_str_scientific(self, n_before, n_decimals, force_dot, group_digits):
         """Put a float in scientific format."""
-        n_work = min(self.digits, n_before + n_decimals)
         if self.is_zero():
-            if not force_dot:
-                if self.exp_sign == 'E':
-                    return 'E+00'
-                return '0D+00'  # matches GW output. odd, odd, odd
-            digitstr = '0' * (n_before + n_decimals)
-            exp10 = 0
-        else:
-            # special case when work_digits == 0, see also below
-            # setting to 0 results in incorrect rounding (why?)
-            mantissa, exp10 = self.to_decimal(1 if n_work == 0 else n_work)
-            digitstr = _get_digits(mantissa, n_work, remove_trailing=True)
-            # append zeros if necessary
-            digitstr = digitstr.ljust(n_decimals + n_before, '0')
+            if force_dot:
+                return b''.join((b'.', (b'0' * n_decimals), self.exp_sign, b'+00'))
+            # single/double difference: this matches GW output. odd, odd, odd
+            if self.exp_sign == b'E':
+                return b'E+00'
+            return b'0D+00'
+        n_work = min(self.digits, n_before + n_decimals)
+        # special case when work_digits == 0, see also below
+        # setting to 0 results in incorrect rounding (why?)
+        mantissa, exp10 = self.to_decimal(1 if n_work == 0 else n_work)
+        digitstr = _get_digits(mantissa, n_work, remove_trailing=True)
+        # append zeros if necessary
+        digitstr = digitstr.ljust(n_decimals + n_before, b'0')
         # this is just to reproduce GW results for no digits:
         # e.g. PRINT USING "#^^^^";1 gives " E+01" not " E+00"
         if n_work == 0:
             exp10 += 1
         exp10 += n_before + n_decimals - 1
-        return self._scientific_notation(digitstr, exp10, digits_to_dot=n_before, force_dot=force_dot, group_digits=group_digits)
+        return self._scientific_notation(
+            digitstr, exp10, digits_to_dot=n_before, force_dot=force_dot, group_digits=group_digits
+        )
 
     def to_str_fixed(self, n_decimals, force_dot, group_digits):
         """Put a float in fixed-point representation."""
+        if self.is_zero():
+            if force_dot:
+                return b'.' + b'0'*n_decimals
+            if n_decimals:
+                return b'0'*n_decimals
+            return b'0'
         # convert to integer_mantissa * 10**exponent
         mantissa, exp10 = self.to_decimal()
         # -exp10 is the number of digits after the radix point
@@ -724,12 +759,14 @@ class Float(Number):
             # this has n_work or n_work+1 digits, depending on rounding
             mantissa, exp10 = self.to_decimal(n_work)
             n_after = -exp10
-        digitstr = str(abs(mantissa))
+        digitstr = b'%d' % (abs(mantissa),)
         # number of digits before the radix point.
         n_before = len(digitstr) - n_after
         # fill up with zeros to required number of figures
-        digitstr = digitstr.ljust(n_decimals + n_before, '0')
-        return self._decimal_notation(digitstr, n_before-1, type_sign='', force_dot=force_dot, group_digits=group_digits)
+        digitstr = digitstr.ljust(n_decimals + n_before, b'0')
+        return self._decimal_notation(
+            digitstr, n_before-1, type_sign=b'', force_dot=force_dot, group_digits=group_digits
+        )
 
     # implementation: floating- and fixed-point decimal notations
 
@@ -739,7 +776,7 @@ class Float(Number):
         chunks = [digitstr[i:i + 3] for i in range(first, len(digitstr), 3)]
         if first:
             chunks = [digitstr[:first]] + chunks
-        return ','.join(chunks)
+        return b','.join(chunks)
 
     def _scientific_notation(self, digitstr, exp10, digits_to_dot, force_dot, group_digits=False):
         """Put digits in scientific E-notation."""
@@ -748,41 +785,37 @@ class Float(Number):
             valstr = self._group_digits(valstr)
         if len(digitstr) > digits_to_dot:
             after_str = digitstr[digits_to_dot:]
-            valstr += '.' + after_str
+            valstr += b'.' + after_str
         elif len(digitstr) == digits_to_dot and force_dot:
-            valstr += '.'
+            valstr += b'.'
         exponent = exp10 - digits_to_dot + 1
         valstr += self.exp_sign
         if exponent < 0:
-            valstr += '-'
+            valstr += b'-'
         else:
-            valstr += '+'
+            valstr += b'+'
         valstr += _get_digits(abs(exponent), n_digits=2, remove_trailing=False)
         return valstr
 
     def _decimal_notation(self, digitstr, exp10, type_sign, force_dot, group_digits=False):
         """Put digits in decimal notation."""
-        type_sign = '' if not type_sign else self.sigil
+        type_sign = b'' if not type_sign else self.sigil
         # digits to decimal point
         exp10 += 1
         if exp10 >= len(digitstr):
-            valstr = digitstr + '0'*(exp10-len(digitstr))
+            valstr = digitstr + b'0'*(exp10-len(digitstr))
             if group_digits:
                 valstr = self._group_digits(valstr)
             if force_dot:
-                valstr += '.'
+                valstr += b'.'
         elif exp10 > 0:
             valstr = digitstr[:exp10]
             if group_digits:
                 valstr = self._group_digits(valstr)
-            valstr += '.' + digitstr[exp10:]
+            valstr += b'.' + digitstr[exp10:]
         else:
-            if force_dot:
-                valstr = '0'
-            else:
-                valstr = ''
-            valstr += '.' + '0'*(-exp10) + digitstr
-        if ('.' not in valstr) or (type_sign == '#'):
+            valstr = b'.' + b'0'*(-exp10) + digitstr
+        if (b'.' not in valstr) or (type_sign == b'#'):
             valstr += type_sign
         return valstr
 
@@ -856,8 +889,10 @@ class Float(Number):
 
     def _denormalise(self):
         """Denormalise to shifted mantissa, exp, sign."""
-        exp = ord(self._buffer[-1])
-        man = struct.unpack(self._intformat, b'\0' + bytearray(self._buffer[:-1]))[0] | self._den_mask
+        exp = bytearray(self._buffer)[-1]
+        man = struct.unpack(
+                self._intformat, b'\0' + bytearray(self._buffer)[:-1]
+            )[0] | self._den_mask
         neg = self.is_negative()
         return exp, man, neg
 
@@ -878,9 +913,11 @@ class Float(Number):
             exp += 1
             man >>= 1
         # pack into byte representation
-        struct.pack_into(self._intformat, self._buffer, 0, (man>>8) & (self._mask if neg else self._posmask))
+        struct.pack_into(
+            self._intformat, self._buffer, 0, (man>>8) & (self._mask if neg else self._posmask)
+        )
         if self._check_limits(exp, neg):
-            self._buffer[-1] = chr(exp)
+            self._buffer[-1:] = int2byte(exp)
         return self
 
     def _to_int_den(self):
@@ -906,7 +943,7 @@ class Float(Number):
             raise OverflowError(self)
         elif exp <= 0:
             # set to zero, but leave mantissa as is
-            self._buffer[-1:] = chr(0)
+            self._buffer[-1:] = int2byte(0)
             return False
         return True
 
@@ -927,13 +964,12 @@ class Float(Number):
             return False
         rhscopy = bytearray(rhs._buffer)
         # so long as the sign is the same ...
-        rhscopy[-2] &= (ord(self._buffer[-2]) | 0x7f)
+        rhscopy[-2] &= (bytearray(self._buffer)[-2] | 0x7f)
         # ... we can compare floats as if they were ints
-        for l, r in reversed(zip(self._buffer, rhscopy)):
-            # memoryview elements are str, bytearray elements are int
-            if ord(l) > r:
+        for l, r in reversed(list(zip(bytearray(self._buffer), bytearray(rhscopy)))):
+            if l > r:
                 return True
-            elif ord(l) < r:
+            elif l < r:
                 return False
         # equal
         return False
@@ -1016,7 +1052,6 @@ class Float(Number):
         return lexp, lman, lneg
 
 
-
 ##############################################################################
 # single-precision floating-point number
 
@@ -1056,7 +1091,7 @@ class Single(Float):
 
     def from_token(self, token):
         """Set value to Single token."""
-        if bytes(token)[0] != tk.T_SINGLE:
+        if bytearray(token)[0] != ord(tk.T_SINGLE):
             raise ValueError('%s is not a Single token.' % repr(token))
         self._buffer[:] = token[-4:]
         return self
@@ -1119,7 +1154,7 @@ class Double(Float):
 
     def from_token(self, token):
         """Set value to Single token."""
-        if bytes(token)[0] != tk.T_DOUBLE:
+        if bytearray(token)[0] != ord(tk.T_DOUBLE):
             raise ValueError('%s is not a Double token.' % repr(token))
         self._buffer[:] = token[-8:]
         return self
@@ -1153,7 +1188,7 @@ def str_to_decimal(s, allow_nonnum=True):
     found_exp_sign, exp_neg, neg = False, False, False
     exp10, exponent, mantissa, digits, zeros = 0, 0, 0, 0, 0
     is_double, is_single = False, False
-    for c in s:
+    for c in iterchar(s):
         # ignore whitespace throughout (x = 1   234  56  .5  means x=123456.5 in gw!)
         if c in BLANKS:
             continue
@@ -1182,7 +1217,7 @@ def str_to_decimal(s, allow_nonnum=True):
                     else:
                         zeros = 0
                 continue
-            elif c == '.':
+            elif c == b'.':
                 found_point = True
                 continue
             elif c.upper() in b'DE':
@@ -1226,8 +1261,8 @@ def str_to_decimal(s, allow_nonnum=True):
 
 def _get_digits(mantissa, n_digits, remove_trailing):
     """Get the digits for an int."""
-    digitstr = str(abs(mantissa)).rjust(n_digits, '0')
+    digitstr = (b'%d' % abs(mantissa)).rjust(n_digits, b'0')
     if remove_trailing:
-        return digitstr.rstrip('0')
+        return digitstr.rstrip(b'0')
     else:
         return digitstr
