@@ -2,15 +2,16 @@
 PC-BASIC - values.py
 Types, values and conversions
 
-(c) 2013--2018 Rob Hagemans
+(c) 2013--2019 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
 import math
-import string
 import struct
 import functools
 
+from ...compat import int2byte
+
 from ..base import error
 from ..base import tokens as tk
 from . import numbers
@@ -32,13 +33,24 @@ TYPE_TO_SIZE = {STR: 3, INT: 2, SNG: 4, DBL: 8}
 SIZE_TO_TYPE = {2: INT, 3: STR, 4: SNG, 8: DBL}
 
 # type classes
-SIZE_TO_CLASS = {2: numbers.Integer, 3: strings.String, 4: numbers.Single, 8: numbers.Double}
-TYPE_TO_CLASS = {INT: numbers.Integer, STR: strings.String, SNG: numbers.Single, DBL: numbers.Double}
+SIZE_TO_CLASS = {
+    2: numbers.Integer,
+    3: strings.String,
+    4: numbers.Single,
+    8: numbers.Double
+}
+
+TYPE_TO_CLASS = {
+    INT: numbers.Integer,
+    STR: strings.String,
+    SNG: numbers.Single,
+    DBL: numbers.Double
+}
 
 
 def size_bytes(name):
     """Return the size of a value type, by variable name or type char."""
-    return TYPE_TO_SIZE[name[-1]]
+    return TYPE_TO_SIZE[name[-1:]]
 
 ###############################################################################
 # type checks
@@ -110,13 +122,18 @@ def _call_float_function(fn, *args):
     feh = args[0].error_handler
     try:
         # to_float can overflow on Double.pos_max
-        args = [arg.to_float(arg._values.double_math) for arg in args]
+        args = [_arg.to_float(values.double_math) for _arg in args]
         floatcls = args[0].__class__
-        args = [arg.to_value() for arg in args]
-        return floatcls(None, values).from_value(fn(*args))
+        args = [_arg.to_value() for _arg in args]
+        result = fn(*args)
+        # python3 may return complex values for some real functions
+        # where python2 simply raises an error
+        if isinstance(result, complex):
+            raise ValueError('Non-real result')
+        return floatcls(None, values).from_value(result)
     except (ValueError, ArithmeticError) as e:
         # create positive infinity of the appropriate class
-        if arg._values.double_math and isinstance(args[0], numbers.Double):
+        if values.double_math and isinstance(args[0], numbers.Double):
             floatcls = numbers.Double
         else:
             floatcls = numbers.Single
@@ -226,7 +243,7 @@ class Values(object):
     def from_bool(self, boo):
         """Convert Python boolean to Integer."""
         if boo:
-            return numbers.Integer(None, self).from_bytes('\xff\xff')
+            return numbers.Integer(None, self).from_bytes(b'\xff\xff')
         return numbers.Integer(None, self)
 
     ###########################################################################
@@ -241,7 +258,7 @@ class Values(object):
         """Convert number token to new Number temporary"""
         if not token:
             raise ValueError('Token must not be empty')
-        lead = bytes(token)[0]
+        lead = bytes(token)[0:1]
         if lead == tk.T_SINGLE:
             return numbers.Single(None, self).from_token(token)
         elif lead == tk.T_DOUBLE:
@@ -260,12 +277,12 @@ class Values(object):
         if typechar == STR:
             return self.new_string().from_str(word)
         # skip spaces and line feeds (but not NUL).
-        word = word.lstrip(' \n').upper()
+        word = word.lstrip(b' \n').upper()
         if not word:
             return self.new_integer()
-        if word[:2] == '&H':
+        if word[:2] == b'&H':
             return self.new_integer().from_hex(word[2:])
-        elif word[:1] == '&':
+        elif word[:1] == b'&':
             return self.new_integer().from_oct(word[2:] if word[1:2] == 'O' else word[1:])
         # we need to try to convert to int first,
         # mainly so that the tokeniser can output the right token type
@@ -349,17 +366,17 @@ def to_int(inp, unsigned=False):
 def mki_(args):
     """MKI$: return the byte representation of an int."""
     x, = args
-    return x._values.new_string().from_str(to_integer(x).to_bytes())
+    return x._values.new_string().from_str(bytes(to_integer(x).to_bytes()))
 
 def mks_(args):
     """MKS$: return the byte representation of a single."""
     x, = args
-    return x._values.new_string().from_str(to_single(x).to_bytes())
+    return x._values.new_string().from_str(bytes(to_single(x).to_bytes()))
 
 def mkd_(args):
     """MKD$: return the byte representation of a double."""
     x, = args
-    return x._values.new_string().from_str(to_double(x).to_bytes())
+    return x._values.new_string().from_str(bytes(to_double(x).to_bytes()))
 
 def cvi_(args):
     """CVI: return the int value of a byte representation."""
@@ -551,7 +568,8 @@ def str_(args):
     """STR$: string representation of a number."""
     x, = args
     return x._values.new_string().from_str(
-                to_repr(pass_number(x), leading_space=True, type_sign=False))
+        to_repr(pass_number(x), leading_space=True, type_sign=False)
+    )
 
 def val_(args):
     """VAL: number value of a string."""
@@ -578,7 +596,7 @@ def chr_(args):
     x, = args
     val = pass_number(x).to_integer().to_int()
     error.range_check(0, 255, val)
-    return x._values.new_string().from_str(chr(val))
+    return x._values.new_string().from_str(int2byte(val))
 
 def oct_(args):
     """OCT$: octal representation of int."""
@@ -656,7 +674,7 @@ def instr_(args):
     new_int = numbers.Integer(None, big._values)
     big = big.to_str()
     small = small.to_str()
-    if big == '' or start > len(big):
+    if big == b'' or start > len(big):
         return new_int
     # BASIC counts string positions from 1
     find = big[start-1:].find(small)
@@ -673,12 +691,12 @@ def string_(args):
         error.range_check(0, 255, asc_value_or_char.to_int())
     list(args)
     if isinstance(asc_value_or_char, strings.String):
-        char = asc_value_or_char.to_str()[0]
+        char = asc_value_or_char.to_str()[:1]
     else:
         # overflow if outside Integer range
         ascval = asc_value_or_char.to_integer().to_int()
         error.range_check(0, 255, ascval)
-        char = chr(ascval)
+        char = int2byte(ascval)
     return strings.String(None, asc_value_or_char._values).from_str(char * num)
 
 ##############################################################################
