@@ -2,34 +2,46 @@
 PC-BASIC - data package
 HEX font loader
 
-(c) 2013--2018 Rob Hagemans
+(c) 2013--2019 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
 import sys
 import pkg_resources
 import logging
+import binascii
+
+from ..compat import iteritems, itervalues, unichr
 
 from ..basic.codepage import PRINTABLE_ASCII
 from .resources import get_data, ResourceFailed
 
+
 FONT_DIR = u'fonts'
 FONT_PATTERN = u'{path}/{name}_{height:02d}.hex'
-FONTS = sorted(set(name.split(u'_', 1)[0] for name in pkg_resources.resource_listdir(__name__, FONT_DIR) if name.lower().endswith(u'.hex')))
+FONTS = sorted(
+    set(name.split(u'_', 1)[0]
+    for name in pkg_resources.resource_listdir(__name__, FONT_DIR)
+    if name.lower().endswith(u'.hex'))
+)
 
 
-def read_fonts(codepage_dict, font_families, warn):
+def read_fonts(codepage_dict, font_families, warn=False):
     """Load font typefaces."""
     # load the graphics fonts, including the 8-pixel RAM font
     # use set() for speed - lookup is O(1) rather than O(n) for list
-    unicode_needed = set(codepage_dict.itervalues())
+    unicode_needed = set(itervalues(codepage_dict))
     # break up any grapheme clusters and add components to set of needed glyphs
     unicode_needed |= set(c for cluster in unicode_needed if len(cluster) > 1 for c in cluster)
     # substitutes is in reverse order: { yen: backslash }
     substitutes = {
         grapheme_cluster: unichr(ord(cp_point))
-        for cp_point, grapheme_cluster in codepage_dict.iteritems()
-        if cp_point in PRINTABLE_ASCII and (len(grapheme_cluster) > 1 or ord(grapheme_cluster) != ord(cp_point))
+        for cp_point, grapheme_cluster in iteritems(codepage_dict)
+        if (
+            cp_point in PRINTABLE_ASCII
+            and (len(grapheme_cluster) > 1
+            or ord(grapheme_cluster) != ord(cp_point))
+        )
     }
     fonts = {}
     # load fonts, height-16 first
@@ -39,12 +51,14 @@ def read_fonts(codepage_dict, font_families, warn):
         for name in font_families:
             try:
                 font_files.append(
-                        get_data(FONT_PATTERN, path=FONT_DIR, name=name, height=height))
+                    get_data(FONT_PATTERN, path=FONT_DIR, name=name, height=height)
+                )
             except ResourceFailed as e:
                 if warn:
                     logging.debug(e)
         fonts[height] = FontLoader(height).load_hex(
-                font_files, unicode_needed, substitutes, warn=warn)
+            font_files, unicode_needed, substitutes, warn=warn
+        )
         # fix missing code points font based on 16-line font
         if fonts[16]:
             fonts[height].fix_missing(unicode_needed, fonts[16])
@@ -52,10 +66,13 @@ def read_fonts(codepage_dict, font_families, warn):
         fonts[9] = fonts[8]
     # convert keys from unicode to codepage
     fonts = {
-        height: {c: font._fontdict[uc] for c, uc in codepage_dict.iteritems() if uc in font._fontdict}
-        for height, font in fonts.iteritems()
+        height: {
+            c: font._fontdict[uc]
+            for c, uc in iteritems(codepage_dict) if uc in font._fontdict
+        }
+        for height, font in iteritems(fonts)
     }
-    return {height: font for height, font in fonts.iteritems()}
+    return {height: font for height, font in iteritems(fonts)}
 
 
 class FontLoader(object):
@@ -75,12 +92,12 @@ class FontLoader(object):
                 continue
             for line in hexres.splitlines():
                 # ignore empty lines and comment lines (first char is #)
-                if (not line) or (line[0] == '#'):
+                if (not line) or (line[:1] == b'#'):
                     continue
                 # strip off comments
                 # split unicodepoint and hex string (max 32 chars)
-                ucs_str, fonthex = line.split('#')[0].split(':')
-                ucs_sequence = ucs_str.split(',')
+                ucs_str, fonthex = line.split(b'#')[0].split(b':')
+                ucs_sequence = ucs_str.split(b',')
                 fonthex = fonthex.strip()
                 # extract codepoint and hex string;
                 # discard anything following whitespace; ignore malformed lines
@@ -100,13 +117,15 @@ class FontLoader(object):
                         fonthex = fonthex[:2*self._height]
                     else:
                         fonthex = fonthex[:4*self._height]
-                    self._fontdict[c] = fonthex.decode('hex')
+                    self._fontdict[c] = binascii.unhexlify(fonthex)
                 except Exception as e:
                     logging.warning('Could not parse line in font file: %s', repr(line))
         # substitute code points
-        self._fontdict.update({old: self._fontdict[new]
-                for (new, old) in substitutes.iteritems()
-                if new in self._fontdict})
+        self._fontdict.update({
+            old: self._fontdict[new]
+            for (new, old) in iteritems(substitutes)
+            if new in self._fontdict
+        })
         # char 0 should always be defined and empty
         self._fontdict[u'\0'] = b'\0' * self._height
         self._combine_glyphs(unicode_needed)
@@ -128,9 +147,10 @@ class FontLoader(object):
                             for y, row in enumerate(self._fontdict[c]):
                                 clusterglyph[y] |= ord(row)
                     except KeyError as e:
-                        logging.debug('Could not combine grapheme cluster %s, missing %s [%s]',
-                            cluster, repr(c), c)
-                    self._fontdict[cluster] = str(clusterglyph)
+                        logging.debug(
+                            'Could not combine grapheme cluster %s, missing %r [%s]', cluster, c, c
+                        )
+                    self._fontdict[cluster] = bytes(clusterglyph)
 
     def _warn_missing(self, unicode_needed, max_warnings=3):
         """Check if we have all needed glyphs."""
@@ -155,8 +175,8 @@ class FontLoader(object):
                     s16 = list(font16.fontdict[c])
                     start = (16 - self._height) // 2
                     if len(s16) == 16:
-                        self._fontdict[c] = ''.join([s16[i] for i in range(start, 16-start)])
+                        self._fontdict[c] = b''.join([s16[i] for i in range(start, 16-start)])
                     else:
-                        self._fontdict[c] = ''.join([s16[i] for i in range(start*2, 32-start*2)])
+                        self._fontdict[c] = b''.join([s16[i] for i in range(start*2, 32-start*2)])
                 except (KeyError, AttributeError) as e:
                     self._fontdict[c] = b'\0' * self._height
