@@ -2,47 +2,153 @@
 PC-BASIC - video_cli.py
 Command-line interface
 
-(c) 2013--2018 Rob Hagemans
+(c) 2013--2019 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
 import sys
 import time
-import threading
-import Queue
 
-from . import ansi
 from .video import VideoPlugin
 from .base import video_plugins, InitFailed
 from ..basic.base import signals
 from ..basic.base import scancode
 from ..basic.base.eascii import as_unicode as uea
-from ..compat import UEOF, console
+from ..compat import EOF, console, stdin, stdout
 
 
 # escape sequence to scancode
-ESC_TO_SCAN = {
-    ansi.F1: scancode.F1,  ansi.F2: scancode.F2,  ansi.F3: scancode.F3,  ansi.F4: scancode.F4,
-    ansi.F1_OLD: scancode.F1,  ansi.F2_OLD: scancode.F2,  ansi.F3_OLD: scancode.F3,
-    ansi.F4_OLD: scancode.F4,  ansi.F5: scancode.F5,  ansi.F6: scancode.F6,  ansi.F7: scancode.F7,
-    ansi.F8: scancode.F8,  ansi.F9: scancode.F9,  ansi.F10: scancode.F10,  ansi.F11: scancode.F11,
-    ansi.F12: scancode.F12,  ansi.END: scancode.END,  ansi.END2: scancode.END,
-    ansi.HOME: scancode.HOME,  ansi.HOME2: scancode.HOME,  ansi.UP: scancode.UP,
-    ansi.DOWN: scancode.DOWN,  ansi.RIGHT: scancode.RIGHT,  ansi.LEFT: scancode.LEFT,
-    ansi.INSERT: scancode.INSERT,  ansi.DELETE: scancode.DELETE,  ansi.PAGEUP: scancode.PAGEUP,
-    ansi.PAGEDOWN: scancode.PAGEDOWN,
-    }
-
+KEY_TO_SCAN = {
+    'F1': scancode.F1,  'F2': scancode.F2,  'F3': scancode.F3,  'F4': scancode.F4,
+    'F5': scancode.F5,  'F6': scancode.F6,  'F7': scancode.F7,  'F8': scancode.F8,
+    'F9': scancode.F9,  'F10': scancode.F10,  'F11': scancode.F11, 'F12': scancode.F12,
+    'END': scancode.END,  'HOME': scancode.HOME,
+    'UP': scancode.UP,  'DOWN': scancode.DOWN,  'RIGHT': scancode.RIGHT,  'LEFT': scancode.LEFT,
+    'INSERT': scancode.INSERT,  'DELETE': scancode.DELETE,  'PAGEUP': scancode.PAGEUP,
+    'PAGEDOWN': scancode.PAGEDOWN,
+}
+MOD_TO_SCAN = {
+    'SHIFT': scancode.LSHIFT,
+    'CTRL': scancode.CTRL,
+    'ALT': scancode.ALT,
+}
 # escape sequence to e-ASCII
-ESC_TO_EASCII = {
-    ansi.F1: uea.F1,  ansi.F2: uea.F2,  ansi.F3: uea.F3,  ansi.F4: uea.F4,  ansi.F1_OLD: uea.F1,
-    ansi.F2_OLD: uea.F2,  ansi.F3_OLD: uea.F3,  ansi.F4_OLD: uea.F4,  ansi.F5: uea.F5,
-    ansi.F6: uea.F6,  ansi.F7: uea.F7,  ansi.F8: uea.F8,  ansi.F9: uea.F9,  ansi.F10: uea.F10,
-    ansi.F11: uea.F11,  ansi.F12: uea.F12,  ansi.END: uea.END,  ansi.END2: uea.END,
-    ansi.HOME: uea.HOME,  ansi.HOME2: uea.HOME,  ansi.UP: uea.UP,  ansi.DOWN: uea.DOWN,
-    ansi.RIGHT: uea.RIGHT,  ansi.LEFT: uea.LEFT,  ansi.INSERT: uea.INSERT,
-    ansi.DELETE: uea.DELETE,  ansi.PAGEUP: uea.PAGEUP,  ansi.PAGEDOWN: uea.PAGEDOWN,
-    }
+KEY_TO_EASCII = {
+    'F1': uea.F1,  'F2': uea.F2,  'F3': uea.F3,  'F4': uea.F4,  'F5': uea.F5,
+    'F6': uea.F6,  'F7': uea.F7,  'F8': uea.F8,  'F9': uea.F9,  'F10': uea.F10,
+    'F11': uea.F11,  'F12': uea.F12,  'END': uea.END,
+    'HOME': uea.HOME,  'UP': uea.UP,  'DOWN': uea.DOWN,
+    'RIGHT': uea.RIGHT,  'LEFT': uea.LEFT,  'INSERT': uea.INSERT,
+    'DELETE': uea.DELETE,  'PAGEUP': uea.PAGEUP,  'PAGEDOWN': uea.PAGEDOWN,
+}
+
+SHIFT_KEY_TO_EASCII = {
+    'F1': uea.SHIFT_F1,
+    'F2': uea.SHIFT_F2,
+    'F3': uea.SHIFT_F3,
+    'F4': uea.SHIFT_F4,
+    'F5': uea.SHIFT_F5,
+    'F6': uea.SHIFT_F6,
+    'F7': uea.SHIFT_F7,
+    'F8': uea.SHIFT_F8,
+    'F9': uea.SHIFT_F9,
+    'F10': uea.SHIFT_F10,
+    'F11': uea.SHIFT_F11,
+    'F12': uea.SHIFT_F12,
+    'HOME': uea.SHIFT_HOME,
+    'UP': uea.SHIFT_UP,
+    'PAGEUP': uea.SHIFT_PAGEUP,
+    'LEFT': uea.SHIFT_LEFT,
+    'RIGHT': uea.SHIFT_RIGHT,
+    'END': uea.SHIFT_END,
+    'DOWN': uea.SHIFT_DOWN,
+    'PAGEDOWN': uea.SHIFT_PAGEDOWN,
+    'INSERT': uea.SHIFT_INSERT,
+    'DELETE': uea.SHIFT_DELETE,
+}
+
+CTRL_KEY_TO_EASCII = {
+    'F1': uea.CTRL_F1,
+    'F2': uea.CTRL_F2,
+    'F3': uea.CTRL_F3,
+    'F4': uea.CTRL_F4,
+    'F5': uea.CTRL_F5,
+    'F6': uea.CTRL_F6,
+    'F7': uea.CTRL_F7,
+    'F8': uea.CTRL_F8,
+    'F9': uea.CTRL_F9,
+    'F10': uea.CTRL_F10,
+    'F11': uea.CTRL_F11,
+    'F12': uea.CTRL_F12,
+    'HOME': uea.CTRL_HOME,
+    'PAGEUP': uea.CTRL_PAGEUP,
+    'LEFT': uea.CTRL_LEFT,
+    'RIGHT': uea.CTRL_RIGHT,
+    'END': uea.CTRL_END,
+    'PAGEDOWN': uea.CTRL_PAGEDOWN,
+    #'ESCAPE': uea.CTRL_ESCAPE,
+    #'BACKSPACE': uea.CTRL_BACKSPACE,
+    #'2': uea.CTRL_2,
+    #'6': uea.CTRL_6,
+    #'MINUS': uea.CTRL_MINUS,
+}
+
+ALT_KEY_TO_EASCII = {
+    '1': uea.ALT_1,
+    '2': uea.ALT_2,
+    '3': uea.ALT_3,
+    '4': uea.ALT_4,
+    '5': uea.ALT_5,
+    '6': uea.ALT_6,
+    '7': uea.ALT_7,
+    '8': uea.ALT_8,
+    '9': uea.ALT_9,
+    '0': uea.ALT_0,
+    '-': uea.ALT_MINUS,
+    '=': uea.ALT_EQUALS,
+    'q': uea.ALT_q,
+    'w': uea.ALT_w,
+    'e': uea.ALT_e,
+    'r': uea.ALT_r,
+    't': uea.ALT_t,
+    'y': uea.ALT_y,
+    'u': uea.ALT_u,
+    'i': uea.ALT_i,
+    'o': uea.ALT_o,
+    'p': uea.ALT_p,
+    'a': uea.ALT_a,
+    's': uea.ALT_s,
+    'd': uea.ALT_d,
+    'f': uea.ALT_f,
+    'g': uea.ALT_g,
+    'h': uea.ALT_h,
+    'j': uea.ALT_j,
+    'k': uea.ALT_k,
+    'l': uea.ALT_l,
+    'z': uea.ALT_z,
+    'x': uea.ALT_x,
+    'c': uea.ALT_c,
+    'v': uea.ALT_v,
+    'b': uea.ALT_b,
+    'n': uea.ALT_n,
+    'm': uea.ALT_m,
+    '\b': uea.ALT_BACKSPACE,
+    '\t': uea.ALT_TAB,
+    '\r': uea.ALT_RETURN,
+    ' ': uea.ALT_SPACE,
+    'F1': uea.ALT_F1,
+    'F2': uea.ALT_F2,
+    'F3': uea.ALT_F3,
+    'F4': uea.ALT_F4,
+    'F5': uea.ALT_F5,
+    'F6': uea.ALT_F6,
+    'F7': uea.ALT_F7,
+    'F8': uea.ALT_F8,
+    'F9': uea.ALT_F9,
+    'F10': uea.ALT_F10,
+    'F11': uea.ALT_F11,
+    'F12': uea.ALT_F12,
+}
 
 
 class VideoTextBase(VideoPlugin):
@@ -51,9 +157,7 @@ class VideoTextBase(VideoPlugin):
     def __init__(self, input_queue, video_queue, **kwargs):
         """Initialise text-based interface."""
         if not console:
-            raise InitFailed('Extension module `win32_console` not compiled.')
-        if not console.is_tty:
-            raise InitFailed('Not a terminal (tty).')
+            raise InitFailed('This interface requires a console terminal (tty).')
         VideoPlugin.__init__(self, input_queue, video_queue)
         # start the stdin thread for non-blocking reads
         self._input_handler = InputHandlerCLI(input_queue)
@@ -124,7 +228,6 @@ class VideoCLI(VideoTextBase):
             if row != self._last_row or col != self._col:
                 self._update_position(row, col)
             console.write(char)
-            #console.flush()
             self._col = (col+2) if is_fullwidth else (col+1)
         # the terminal cursor has moved, so we'll need to move it back later
         # if that's not where we want to be
@@ -145,8 +248,7 @@ class VideoCLI(VideoTextBase):
         if (self._vpagenum == self._apagenum and
                 start <= self._cursor_row and stop >= self._cursor_row):
             self._update_position(self._cursor_row, 1)
-            console.write(ansi.CLEAR_LINE.decode('ascii'))
-            #console.flush()
+            console.clear_row()
 
     def scroll_up(self, from_line, scroll_height, back_attr):
         """Scroll the screen up between from_line and scroll_height."""
@@ -157,7 +259,6 @@ class VideoCLI(VideoTextBase):
         if self._vpagenum != self._apagenum:
             return
         console.write(u'\r\n')
-        #console.flush()
 
     def scroll_down(self, from_line, scroll_height, back_attr):
         """Scroll the screen down between from_line and scroll_height."""
@@ -192,7 +293,6 @@ class VideoCLI(VideoTextBase):
         rowtext = (u''.join(self._text[self._vpagenum][row-1]))
         console.write(rowtext.replace(u'\0', u' '))
         self._col = len(self._text[self._vpagenum][row-1])+1
-        #console.flush()
 
     def _update_position(self, row, col):
         """Move terminal print location."""
@@ -200,7 +300,6 @@ class VideoCLI(VideoTextBase):
         if row and row != self._last_row:
             if self._last_row:
                 console.write(u'\r\n')
-                #console.flush()
                 self._col = 1
             self._last_row = row
             # show what's on the line where we are.
@@ -211,11 +310,9 @@ class VideoCLI(VideoTextBase):
         """Move terminal print column."""
         if col != self._col:
             if self._col > col:
-                console.write(ansi.MOVE_N_LEFT.decode('ascii') % (self._col-col))
-                #console.flush()
+                console.move_cursor_left(self._col-col)
             elif self._col < col:
-                console.write(ansi.MOVE_N_RIGHT.decode('ascii') % (col-self._col))
-                #console.flush()
+                console.move_cursor_right(col-self._col)
             self._col = col
 
 
@@ -228,24 +325,26 @@ class InputHandlerCLI(object):
         """Start the keyboard reader."""
         self._input_queue = queue
         self._f12_active = False
+        self.quit_on_eof = True
 
     def drain_queue(self):
         """Handle keyboard events."""
         while True:
             # s is one unicode char or one scancode
-            uc, sc = self._get_key()
+            uc, sc, mods = self._get_key()
             if not uc and not sc:
                 break
-            if uc == UEOF:
+            if uc == EOF and self.quit_on_eof:
                 # ctrl-D (unix) / ctrl-Z (windows)
                 self._input_queue.put(signals.Event(signals.KEYB_QUIT))
             elif uc == u'\x7f':
                 # backspace
                 self._input_queue.put(
-                        signals.Event(signals.KEYB_DOWN, (uea.BACKSPACE, scancode.BACKSPACE, [])))
+                    signals.Event(signals.KEYB_DOWN, (uea.BACKSPACE, scancode.BACKSPACE, []))
+                )
             elif sc or uc:
                 # check_full=False to allow pasting chunks of text
-                self._input_queue.put(signals.Event(signals.KEYB_DOWN, (uc, sc, [])))
+                self._input_queue.put(signals.Event(signals.KEYB_DOWN, (uc, sc, mods)))
                 # this is needed since we don't send key-up events at all otherwise
                 if sc == scancode.F12:
                     self._f12_active = True
@@ -254,32 +353,21 @@ class InputHandlerCLI(object):
                     self._f12_active = False
 
     def _get_key(self):
-        """Retrieve one scancode sequence or one unicode char from keyboard."""
-        s = console.read_char()
-        if s == u'':
-            return None, None
-        # ansi sequences start with \x1b
-        esc = (s == ansi.ESC)
-        # escape sequences are at most 5 chars long
-        more = 5
-        cutoff = 100
-        if not esc:
-            return s, None
-        while (more > 0) and (cutoff > 0):
-            if esc:
-                # return the first recognised escape sequence
-                uc = ESC_TO_EASCII.get(s, u'')
-                scan = ESC_TO_SCAN.get(s, None)
-                if uc or scan:
-                    return uc, scan
-            # give time for the queue to fill up
-            time.sleep(0.0005)
-            c = console.read_char()
-            cutoff -= 1
-            if c == u'':
-                continue
-            more -= 1
-            s += c
-        # no sequence or decodable string found
-        # decode as good as it gets
-        return s, None
+        """Retrieve one keypress."""
+        char, key, mods = console.read_key()
+        if not char and not key:
+            return None, None, []
+        # override characters wth alt
+        if 'ALT' in mods and 'CTRL' not in mods:
+            char = ALT_KEY_TO_EASCII.get(key, u'')
+        # don't override ctrl characters
+        if not char:
+            if 'CTRL' in mods and 'ALT' not in mods:
+                char = CTRL_KEY_TO_EASCII.get(key, u'')
+            elif mods == {'SHIFT'}:
+                char = SHIFT_KEY_TO_EASCII.get(key, char)
+            else:
+                char = KEY_TO_EASCII.get(key, char)
+        scan = KEY_TO_SCAN.get(key, None)
+        modscan = [MOD_TO_SCAN[mod] for mod in mods if mod in MOD_TO_SCAN]
+        return char, scan, modscan
