@@ -2,14 +2,18 @@
 PC-BASIC - video_pygame.py
 Graphical interface based on PyGame
 
-(c) 2013--2018 Rob Hagemans
+(c) 2013--2019 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
+import sys
 import logging
 
+from ..compat import iteritems, unichr, muffle
+
 try:
-    import pygame
+    with muffle(sys.stdout):
+        import pygame
 except ImportError:
     pygame = None
 
@@ -18,11 +22,13 @@ try:
 except ImportError:
     numpy = None
 
+
 from ..basic.base import signals
 from ..basic.base import scancode
 from ..basic.base.eascii import as_unicode as uea
 from ..data.resources import ICON
-from ..compat import WIN32, MACOS
+from ..compat import WIN32, MACOS, PY2
+from ..compat import set_dpi_aware
 from .video import VideoPlugin
 from .base import video_plugins, InitFailed, EnvironmentCache, NOKILL_MESSAGE
 from . import clipboard
@@ -46,9 +52,12 @@ class VideoPygame(VideoPlugin):
             caption=u'', icon=ICON,
             scaling=None, dimensions=None, aspect_ratio=(4, 3), border_width=0, fullscreen=False,
             prevent_close=False, mouse_clipboard=True,
-            **kwargs):
+            **kwargs
+        ):
         """Initialise pygame interface."""
+        logging.warning('The PyGame interface is deprecated, use the SDL2 interface instead.')
         VideoPlugin.__init__(self, input_queue, video_queue)
+
         # request smooth scaling
         self._smooth = scaling == 'smooth'
         # ignore ALT+F4 and window X button
@@ -63,6 +72,8 @@ class VideoPygame(VideoPlugin):
             raise InitFailed('Module `pygame` not found')
         if not numpy:
             raise InitFailed('Module `numpy` not found')
+        # Windows 10 - set to DPI aware to avoid scaling twice on HiDPI screens
+        set_dpi_aware()
         # ensure we have the correct video driver for SDL 1.2
         # pygame sets this on import, but if we've tried SDL2 we've had to
         # reset this value
@@ -82,13 +93,12 @@ class VideoPygame(VideoPlugin):
         # border attribute
         self.border_attr = 0
         # palette and colours
-        # composite colour artifacts
-        self._composite = False
         # working palette - attribute index in blue channel
         self.work_palette = [(0, 0, index) for index in range(256)]
         # display palettes for blink states 0, 1
         self._palette = [None, None]
-        self._saved_palette = [None, None]
+        # composite colour artifacts
+        self._pixel_packing = False
         # text attributes supported
         self.mode_has_blink = True
         # update cycle
@@ -116,25 +126,30 @@ class VideoPygame(VideoPlugin):
         # get physical screen dimensions (needs to be called before set_mode)
         display_info = pygame.display.Info()
         self._window_sizer = window.WindowSizer(
-                display_info.current_w, display_info.current_h,
-                scaling, dimensions, aspect_ratio, border_width, fullscreen)
+            display_info.current_w, display_info.current_h,
+            scaling, dimensions, aspect_ratio, border_width,
+        )
         # determine initial display size
-        self.display_size = self._window_sizer.find_display_size(720, 400)
+        self._window_sizer.set_canvas_size(720, 400)
         self._set_icon(icon)
         try:
-            self._resize_display(*self.display_size)
+            self._resize_display()
         except pygame.error as e:
             self._close_pygame()
             raise InitFailed('Could not initialise display: %s' % e)
         if self._smooth and self.display.get_bitsize() < 24:
             logging.warning(
-                    'Smooth scaling not available on this display (depth %d < 24)',
-                    self.display.get_bitsize())
+                'Smooth scaling not available on this display (depth %d < 24)',
+                self.display.get_bitsize()
+            )
             self._smooth = False
-        pygame.display.set_caption(self.caption.encode('utf-8', 'replace'))
+        if PY2:
+            pygame.display.set_caption(self.caption.encode('utf-8', 'replace'))
+        else:
+            pygame.display.set_caption(self.caption)
         pygame.key.set_repeat(500, 24)
         # load an all-black 16-colour game palette to get started
-        self.set_palette([(0,0,0)]*16, None)
+        self.set_palette([(0,0,0)]*16, None, None)
         pygame.joystick.init()
         self.joysticks = [pygame.joystick.Joystick(x) for x in range(pygame.joystick.get_count())]
         for j in self.joysticks:
@@ -199,40 +214,51 @@ class VideoPygame(VideoPlugin):
                     if event.button == 1:
                         # LEFT button: copy
                         pos = self._window_sizer.normalise_pos(*event.pos)
-                        self.clipboard.start(1 + pos[1] // self.font_height,
-                                1 + (pos[0]+self.font_width//2) // self.font_width)
+                        self.clipboard.start(
+                            1 + pos[1] // self.font_height,
+                            1 + (pos[0]+self.font_width//2) // self.font_width
+                        )
                     elif event.button == 2:
                         # MIDDLE button: paste
                         text = self.clipboard_handler.paste(mouse=True)
                         self.clipboard.paste(text)
+                    self.busy = True
                 if event.button == 1:
                     # right mouse button is a pen press
-                    self._input_queue.put(signals.Event(signals.PEN_DOWN,
-                                                self._window_sizer.normalise_pos(*event.pos)))
+                    self._input_queue.put(signals.Event(
+                        signals.PEN_DOWN, self._window_sizer.normalise_pos(*event.pos)
+                    ))
             elif event.type == pygame.MOUSEBUTTONUP:
                 self._input_queue.put(signals.Event(signals.PEN_UP))
                 if self._mouse_clip and event.button == 1:
                     self.clipboard.copy(mouse=True)
                     self.clipboard.stop()
+                    self.busy = True
             elif event.type == pygame.MOUSEMOTION:
                 pos = self._window_sizer.normalise_pos(*event.pos)
                 self._input_queue.put(signals.Event(signals.PEN_MOVED, pos))
                 if self.clipboard.active():
-                    self.clipboard.move(1 + pos[1] // self.font_height,
-                           1 + (pos[0]+self.font_width//2) // self.font_width)
+                    self.clipboard.move(
+                        1 + pos[1] // self.font_height,
+                        1 + (pos[0]+self.font_width//2) // self.font_width
+                    )
+                    self.busy = True
             elif event.type == pygame.JOYBUTTONDOWN:
-                self._input_queue.put(signals.Event(signals.STICK_DOWN,
-                                                      (event.joy, event.button)))
+                self._input_queue.put(signals.Event(
+                    signals.STICK_DOWN, (event.joy, event.button)
+                ))
             elif event.type == pygame.JOYBUTTONUP:
-                self._input_queue.put(signals.Event(signals.STICK_UP,
-                                                      (event.joy, event.button)))
+                self._input_queue.put(signals.Event(
+                    signals.STICK_UP, (event.joy, event.button)
+                ))
             elif event.type == pygame.JOYAXISMOTION:
-                self._input_queue.put(signals.Event(signals.STICK_MOVED,
-                                                      (event.joy, event.axis,
-                                                      int(event.value*127 + 128))))
+                self._input_queue.put(signals.Event(
+                    signals.STICK_MOVED, (event.joy, event.axis, int(event.value*127 + 128))
+                ))
             elif event.type == pygame.VIDEORESIZE:
                 if not self.fullscreen:
-                    self._resize_display(event.w, event.h)
+                    self._window_sizer.set_display_size(event.w, event.h)
+                    self._resize_display()
             elif event.type == pygame.QUIT:
                 if self._nokill:
                     self.set_caption_message(NOKILL_MESSAGE)
@@ -242,27 +268,28 @@ class VideoPygame(VideoPlugin):
     def _handle_key_down(self, e):
         """Handle key-down event."""
         # get scancode
-        scan = key_to_scan.get(e.key, None)
+        scan = KEY_TO_SCAN.get(e.key, None)
         # get modifiers
-        mod = [s for m, s in mod_to_scan.iteritems() if e.mod & m]
+        mod = [s for m, s in iteritems(MOD_TO_SCAN) if e.mod & m]
         # compensate for missing l-alt down events (needed at least on Ubuntu Unity)
         if (e.mod & pygame.KMOD_LALT) and not self._alt_key_down:
             # insert an alt keydown event before the alt-modified keydown
             fake_event = pygame.event.Event(
-                    pygame.KEYDOWN, scancode=0, key=pygame.K_LALT, unicode=u'', mod=0)
+                pygame.KEYDOWN, scancode=0, key=pygame.K_LALT, unicode=u'', mod=0
+            )
             self._handle_key_down(fake_event)
         if e.key == pygame.K_LALT:
             self._alt_key_down = True
         # get eascii
         try:
             if e.mod & pygame.KMOD_LALT:
-                c = alt_key_to_eascii[e.key]
+                c = ALT_KEY_TO_EASCII[e.key]
             elif e.mod & pygame.KMOD_CTRL:
-                c = ctrl_key_to_eascii[e.key]
+                c = CTRL_KEY_TO_EASCII[e.key]
             elif e.mod & pygame.KMOD_SHIFT:
-                c = shift_key_to_eascii[e.key]
+                c = SHIFT_KEY_TO_EASCII[e.key]
             else:
-                c = key_to_eascii[e.key]
+                c = KEY_TO_EASCII[e.key]
         except KeyError:
             key = e.key
             if e.mod & pygame.KMOD_CTRL and key >= ord('a') and key <= ord('z'):
@@ -279,7 +306,8 @@ class VideoPygame(VideoPlugin):
             # F11+f to toggle fullscreen mode
             if e.key == pygame.K_f:
                 self.fullscreen = not self.fullscreen
-                self._resize_display(*self._window_sizer.find_display_size(*self.size))
+                self._window_sizer.set_canvas_size(*self.size, fullscreen=self.fullscreen)
+                self._resize_display()
             self.clipboard.handle_key(scan, c)
             self.busy = True
         else:
@@ -289,12 +317,11 @@ class VideoPygame(VideoPlugin):
             # fix missing ascii for numeric keypad on Windows
             if e.mod & pygame.KMOD_NUM:
                 try:
-                    c = key_to_eascii_num[e.key]
+                    c = KEY_TO_EASCII_NUM[e.key]
                 except KeyError:
                     pass
             # insert into keyboard queue
-            self._input_queue.put(signals.Event(
-                                    signals.KEYB_DOWN, (c, scan, mod)))
+            self._input_queue.put(signals.Event(signals.KEYB_DOWN, (c, scan, mod)))
 
     def _handle_key_up(self, e):
         """Handle key-up event."""
@@ -302,7 +329,8 @@ class VideoPygame(VideoPlugin):
         if (e.key == pygame.K_LALT or e.mod & pygame.KMOD_LALT) and not self._alt_key_down:
             # insert an alt keydown event before the alt-modified keydown
             fake_event = pygame.event.Event(
-                    pygame.KEYDOWN, scancode=0, key=pygame.K_LALT, unicode=u'', mod=0)
+                pygame.KEYDOWN, scancode=0, key=pygame.K_LALT, unicode=u'', mod=0
+            )
             self._handle_key_down(fake_event)
             if e.key != pygame.K_LALT:
                 # insert a keydown before the keyup
@@ -317,8 +345,7 @@ class VideoPygame(VideoPlugin):
             self.f11_active = False
         # last key released gets remembered
         try:
-            self._input_queue.put(signals.Event(
-                                    signals.KEYB_UP, (key_to_scan[e.key],)))
+            self._input_queue.put(signals.Event(signals.KEYB_UP, (KEY_TO_SCAN[e.key],)))
         except KeyError:
             pass
 
@@ -340,7 +367,8 @@ class VideoPygame(VideoPlugin):
             if self._cycle % BLINK_CYCLES == 0:
                 self.busy = True
         if self.cursor_visible and (
-                (self.cursor_row != self.last_row) or (self.cursor_col != self.last_col)):
+                (self.cursor_row != self.last_row) or (self.cursor_col != self.last_col)
+            ):
             self.busy = True
         tock = pygame.time.get_ticks()
         if tock - self.last_cycle >= CYCLE_TIME:
@@ -355,12 +383,13 @@ class VideoPygame(VideoPlugin):
     def _do_flip(self):
         """Draw the canvas to the screen."""
         # create the screen that will be stretched onto the display
-        border_x, border_y = self._window_sizer.border_start()
+        border_x, border_y = self._window_sizer.border_shift
         # surface depth and flags match those of canvas
         # pylint: disable=E1121,E1123
         screen = pygame.Surface(
             (self.size[0] + 2*border_x, self.size[1] + 2*border_y),
-            0, self.canvas[self.vpagenum])
+            0, self.canvas[self.vpagenum]
+        )
         screen.set_palette(self.work_palette)
         # border colour
         border_colour = pygame.Color(0, 0, self.border_attr % self.num_fore_attrs)
@@ -371,15 +400,24 @@ class VideoPygame(VideoPlugin):
         self._draw_cursor(workscreen)
         if self.clipboard.active():
             create_feedback(workscreen, self.clipboard.selection_rect)
-        if self._composite:
-            screen = apply_composite_artifacts(screen, 4//self.bitsperpixel)
+        if self._pixel_packing:
+            screen = apply_composite_artifacts(screen, *self._pixel_packing)
         screen.set_palette(self._palette[self.blink_state])
+        letterbox = pygame.Rect(
+            self._window_sizer.letterbox_shift, self._window_sizer.window_size
+        )
         if self._smooth:
-            pygame.transform.smoothscale(screen.convert(self.display),
-                                         self.display.get_size(), self.display)
+            # smoothscale to subsurface does not work correctly, so create a new surface and blit
+            dest_surf = pygame.Surface(self._window_sizer.window_size) #, self.display)
+            pygame.transform.smoothscale(
+                screen.convert(dest_surf), self._window_sizer.window_size, dest_surf
+            )
+            self.display.blit(dest_surf, letterbox)
         else:
-            pygame.transform.scale(screen.convert(self.display),
-                                   self.display.get_size(), self.display)
+            dest_surf = self.display.subsurface(letterbox)
+            pygame.transform.scale(
+                screen.convert(dest_surf), self._window_sizer.window_size, dest_surf
+            )
         pygame.display.flip()
 
     def _draw_cursor(self, screen):
@@ -388,25 +426,31 @@ class VideoPygame(VideoPlugin):
             return
         # copy screen under cursor
         self.under_top_left = (
-                (self.cursor_col-1) * self.font_width, (self.cursor_row-1) * self.font_height)
+            (self.cursor_col-1) * self.font_width,
+            (self.cursor_row-1) * self.font_height
+        )
         under_char_area = pygame.Rect(
-                (self.cursor_col-1) * self.font_width, (self.cursor_row-1) * self.font_height,
-                self.cursor_width, self.font_height)
+            (self.cursor_col-1) * self.font_width, (self.cursor_row-1) * self.font_height,
+            self.cursor_width, self.font_height
+        )
         if self.text_mode:
             # cursor is visible - to be done every cycle between 5 and 10, 15 and 20
-            if self._cycle / BLINK_CYCLES in (1, 3):
-                screen.blit(self.cursor, (
-                        (self.cursor_col-1) * self.font_width,
-                        (self.cursor_row-1) * self.font_height) )
+            if self._cycle // BLINK_CYCLES in (1, 3):
+                screen.blit(
+                    self.cursor,
+                    ((self.cursor_col-1) * self.font_width, (self.cursor_row-1) * self.font_height)
+                )
         else:
             index = self.cursor_attr % self.num_fore_attrs
             # reference the destination area
             dest_array = pygame.surfarray.pixels2d(
-                    screen.subsurface(pygame.Rect(
-                        (self.cursor_col-1) * self.font_width,
-                        (self.cursor_row-1) * self.font_height + self.cursor_from,
-                        self.cursor_width,
-                        self.cursor_to - self.cursor_from + 1)))
+                screen.subsurface(pygame.Rect(
+                    (self.cursor_col-1) * self.font_width,
+                    (self.cursor_row-1) * self.font_height + self.cursor_from,
+                    self.cursor_width,
+                    self.cursor_to - self.cursor_from + 1
+                ))
+            )
             dest_array ^= index
         self.last_row = self.cursor_row
         self.last_col = self.cursor_col
@@ -414,14 +458,13 @@ class VideoPygame(VideoPlugin):
     ###########################################################################
     # miscellaneous helper functions
 
-    def _resize_display(self, width, height):
+    def _resize_display(self):
         """Change the display size."""
         if self.fullscreen:
-            flags = pygame.FULLSCREEN | pygame.NOFRAME
+            flags = pygame.NOFRAME
         else:
             flags = pygame.RESIZABLE
-        self.display = pygame.display.set_mode((width, height), flags)
-        self._window_sizer.window_size = width, height
+        self.display = pygame.display.set_mode(self._window_sizer.display_size, flags)
         # load display if requested
         self.busy = True
 
@@ -437,24 +480,24 @@ class VideoPygame(VideoPlugin):
         self.font_width = mode_info.font_width
         self.num_pages = mode_info.num_pages
         self.mode_has_blink = mode_info.has_blink
-        if not self.text_mode:
-            self.bitsperpixel = mode_info.bitsperpixel
         # logical size
         self.size = (mode_info.pixel_width, mode_info.pixel_height)
-        self._window_sizer.size = self.size
-        self._resize_display(*self._window_sizer.find_display_size(*self.size))
+        self._window_sizer.set_canvas_size(*self.size, fullscreen=self.fullscreen)
+        self._resize_display()
         # set standard cursor
-        self.set_cursor_shape(self.font_width, self.font_height,
-                              0, self.font_height)
+        self.set_cursor_shape(self.font_width, self.font_height, 0, self.font_height)
         # whole screen (blink on & off)
-        self.canvas = [ pygame.Surface(self.size, depth=8) # pylint: disable=E1121,E1123
-                        for _ in range(self.num_pages)]
+        self.canvas = [
+            pygame.Surface(self.size, depth=8) # pylint: disable=E1121,E1123
+            for _ in range(self.num_pages)
+        ]
         for i in range(self.num_pages):
             self.canvas[i].set_palette(self.work_palette)
         # initialise clipboard
         self.clipboard = clipboard.ClipboardInterface(
-                self.clipboard_handler, self._input_queue,
-                mode_info.width, mode_info.height, self.font_width, self.font_height, self.size)
+            self.clipboard_handler, self._input_queue,
+            mode_info.width, mode_info.height, self.font_width, self.font_height, self.size
+        )
         self.busy = True
         self._has_window = True
 
@@ -467,7 +510,7 @@ class VideoPygame(VideoPlugin):
         """Put text on the clipboard."""
         self.clipboard_handler.copy(text, mouse)
 
-    def set_palette(self, rgb_palette_0, rgb_palette_1):
+    def set_palette(self, rgb_palette_0, rgb_palette_1, pack_pixels):
         """Build the palette."""
         self.num_fore_attrs = min(16, len(rgb_palette_0))
         self.num_back_attrs = min(8, self.num_fore_attrs)
@@ -477,9 +520,12 @@ class VideoPygame(VideoPlugin):
         # bottom 128 are non-blink, top 128 blink to background
         self._palette[0] = rgb_palette_0[:self.num_fore_attrs] * (256//self.num_fore_attrs)
         self._palette[1] = rgb_palette_1[:self.num_fore_attrs] * (128//self.num_fore_attrs)
-        for b in rgb_palette_1[:self.num_back_attrs] * (
-                    128 // self.num_fore_attrs // self.num_back_attrs):
+        for b in (
+                rgb_palette_1[:self.num_back_attrs] *
+                (128 // self.num_fore_attrs // self.num_back_attrs)
+            ):
             self._palette[1] += [b]*self.num_fore_attrs
+        self._pixel_packing = pack_pixels
         self.busy = True
 
     def set_border_attr(self, attr):
@@ -487,20 +533,12 @@ class VideoPygame(VideoPlugin):
         self.border_attr = attr
         self.busy = True
 
-    def set_composite(self, on, composite_colors):
-        """Enable/disable composite artifacts."""
-        if on != self._composite:
-            self._palette, self._saved_palette = self._saved_palette, self._palette
-        if on:
-            self._palette = [composite_colors] * 2
-        self._composite = on
-        self.busy = True
-
     def clear_rows(self, back_attr, start, stop):
         """Clear a range of screen rows."""
         bg = (0, 0, back_attr)
         scroll_area = pygame.Rect(
-                0, (start-1)*self.font_height, self.size[0], (stop-start+1)*self.font_height)
+            0, (start-1)*self.font_height, self.size[0], (stop-start+1)*self.font_height
+        )
         self.canvas[self.apagenum].fill(bg, scroll_area)
         self.busy = True
 
@@ -531,29 +569,33 @@ class VideoPygame(VideoPlugin):
     def scroll_up(self, from_line, scroll_height, back_attr):
         """Scroll the screen up between from_line and scroll_height."""
         temp_scroll_area = pygame.Rect(
-                0, (from_line-1)*self.font_height,
-                self.size[0], (scroll_height-from_line+1) * self.font_height)
+            0, (from_line-1)*self.font_height,
+            self.size[0], (scroll_height-from_line+1) * self.font_height
+        )
         # scroll
         self.canvas[self.apagenum].set_clip(temp_scroll_area)
         self.canvas[self.apagenum].scroll(0, -self.font_height)
         # empty new line
         bg = (0, 0, back_attr)
-        self.canvas[self.apagenum].fill(bg, (
-                0, (scroll_height-1) * self.font_height, self.size[0], self.font_height))
+        self.canvas[self.apagenum].fill(
+            bg, (0, (scroll_height-1) * self.font_height, self.size[0], self.font_height)
+        )
         self.canvas[self.apagenum].set_clip(None)
         self.busy = True
 
     def scroll_down(self, from_line, scroll_height, back_attr):
         """Scroll the screen down between from_line and scroll_height."""
         temp_scroll_area = pygame.Rect(
-                0, (from_line-1) * self.font_height,
-                self.size[0], (scroll_height-from_line+1) * self.font_height)
+            0, (from_line-1) * self.font_height,
+            self.size[0], (scroll_height-from_line+1) * self.font_height
+        )
         self.canvas[self.apagenum].set_clip(temp_scroll_area)
         self.canvas[self.apagenum].scroll(0, self.font_height)
         # empty new line
         bg = (0, 0, back_attr)
-        self.canvas[self.apagenum].fill(bg, (
-                0, (from_line-1) * self.font_height, self.size[0], self.font_height))
+        self.canvas[self.apagenum].fill(
+            bg, (0, (from_line-1) * self.font_height, self.size[0], self.font_height)
+        )
         self.canvas[self.apagenum].set_clip(None)
         self.busy = True
 
@@ -565,18 +607,18 @@ class VideoPygame(VideoPlugin):
         color = (0, 0, fore + self.num_fore_attrs*back + 128*blink)
         bg = (0, 0, back)
         x0, y0 = (col-1)*self.font_width, (row-1)*self.font_height
-        if cp == '\0':
+        if cp == u'\0':
             # guaranteed to be blank, saves time on some BLOADs
             self.canvas[pagenum].fill(bg, (x0, y0, self.font_width, self.font_height))
         else:
             try:
                 glyph = self.glyph_dict[cp]
             except KeyError:
-                if '\0' not in self.glyph_dict:
+                if u'\0' not in self.glyph_dict:
                     logging.error('No glyph received for code point 0')
                     return
                 logging.warning('No glyph received for code point %s', hex(ord(cp)))
-                glyph = self.glyph_dict['\0']
+                glyph = self.glyph_dict[u'\0']
             if glyph.get_palette_at(0) != bg:
                 glyph.set_palette_at(0, bg)
             if glyph.get_palette_at(1) != color:
@@ -588,7 +630,7 @@ class VideoPygame(VideoPlugin):
 
     def build_glyphs(self, new_dict):
         """Build a dict of glyphs for use in text mode."""
-        for char, glyph in new_dict.iteritems():
+        for char, glyph in iteritems(new_dict):
             self.glyph_dict[char] = glyph_to_surface(glyph)
 
     def set_cursor_shape(self, width, height, from_line, to_line):
@@ -622,8 +664,9 @@ class VideoPygame(VideoPlugin):
     def put_interval(self, pagenum, x, y, colours):
         """Write a list of attributes to a scanline interval."""
         # reference the interval on the canvas
-        pygame.surfarray.pixels2d(self.canvas[pagenum]
-                )[x:x+len(colours), y] = numpy.array(colours).astype(int)
+        pygame.surfarray.pixels2d(self.canvas[pagenum])[x:x+len(colours), y] = (
+            numpy.array(colours).astype(int)
+        )
         self.busy = True
 
     def put_rect(self, pagenum, x0, y0, x1, y1, array):
@@ -631,8 +674,11 @@ class VideoPygame(VideoPlugin):
         if (x1 < x0) or (y1 < y0):
             return
         # reference the destination area
-        pygame.surfarray.pixels2d(self.canvas[pagenum].subsurface(
-            pygame.Rect(x0, y0, x1-x0+1, y1-y0+1)))[:] = numpy.array(array).T
+        pygame.surfarray.pixels2d(
+            self.canvas[pagenum].subsurface(
+                pygame.Rect(x0, y0, x1-x0+1, y1-y0+1)
+            )
+        )[:] = numpy.array(array).T
         self.busy = True
 
 
@@ -643,8 +689,10 @@ class PygameClipboard(clipboard.Clipboard):
     """Clipboard handling using Pygame.Scrap."""
 
     # text type we look for in the clipboard
-    text = ('UTF8_STRING', 'text/plain;charset=utf-8', 'text/plain',
-            'TEXT', 'STRING')
+    text = (
+        'UTF8_STRING', 'text/plain;charset=utf-8', 'text/plain',
+        'TEXT', 'STRING'
+    )
 
     def __init__(self):
         """Initialise the clipboard handler."""
@@ -667,11 +715,12 @@ class PygameClipboard(clipboard.Clipboard):
                 # on Windows, encode as utf-16 without FF FE byte order mark and null-terminate
                 # but give it a utf-8 MIME type, because that's how Windows likes it
                 pygame.scrap.put(
-                        'text/plain;charset=utf-8', text.encode('utf-16le', 'replace') + '\0\0')
+                    'text/plain;charset=utf-8', text.encode('utf-16le', 'replace') + b'\0\0'
+                )
             else:
                 pygame.scrap.put(pygame.SCRAP_TEXT, text.encode('utf-8', 'replace'))
-        except KeyError:
-            logging.debug('Clipboard copy failed for clip %s', repr(text))
+        except Exception as e:# KeyError:
+            logging.debug('Clipboard copy failed for clip %r: %s', text, e)
 
     def paste(self, mouse=False):
         """Return unicode text from clipboard."""
@@ -679,34 +728,40 @@ class PygameClipboard(clipboard.Clipboard):
             pygame.scrap.set_mode(pygame.SCRAP_SELECTION)
         else:
             pygame.scrap.set_mode(pygame.SCRAP_CLIPBOARD)
-        us = ''
+        us = u''
+        s = b''
         available = pygame.scrap.get_types()
         for text_type in self.text:
             if text_type not in available:
                 continue
-            us = pygame.scrap.get(text_type)
-            if us:
+            s = pygame.scrap.get(text_type)
+            if s:
                 break
         if WIN32:
             if text_type == 'text/plain;charset=utf-8':
                 # it's lying, it's giving us UTF16 little-endian
                 # ignore any bad UTF16 characters from outside
-                us = us.decode('utf-16le', errors='replace')
+                us = s.decode('utf-16le', errors='replace')
+            else:
+                # fallback
+                us = s.decode('ascii', errors='replace')
             # remove null-terminator
             us = us[:us.find(u'\0')]
         else:
-            us = us.decode('utf-8', errors='replace')
+            us = s.decode('utf-8', errors='replace')
         if us:
-            us = us.replace('\r\n', '\n').replace('\n', '\r')
+            us = us.replace(u'\r\n', u'\n').replace(u'\n', u'\r')
         return us or u''
 
 def get_clipboard_handler():
     """Get a working Clipboard handler object."""
-    # Pygame.Scrap doesn't work on OSX
+    # Pygame.Scrap doesn't work on OSX and crashes on Linux
     if MACOS:
         handler = clipboard.MacClipboard()
-    else:
+    elif WIN32:
         handler = PygameClipboard()
+    else:
+        handler = clipboard.XClipboard()
     if not handler.ok:
         logging.warning('Clipboard copy and paste not available.')
         handler = clipboard.Clipboard()
@@ -727,7 +782,7 @@ def create_feedback(surface, selection_rects):
 
 if pygame:
     # these are PC keyboard scancodes
-    key_to_scan = {
+    KEY_TO_SCAN = {
         # top row
         pygame.K_ESCAPE: scancode.ESCAPE, pygame.K_1: scancode.N1,
         pygame.K_2: scancode.N2, pygame.K_3: scancode.N3,
@@ -795,8 +850,7 @@ if pygame:
         pygame.K_BREAK: scancode.BREAK,
     }
 
-
-    key_to_eascii = {
+    KEY_TO_EASCII = {
         pygame.K_F1: uea.F1,
         pygame.K_F2: uea.F2,
         pygame.K_F3: uea.F3,
@@ -827,7 +881,7 @@ if pygame:
         pygame.K_DELETE: uea.DELETE,
     }
 
-    shift_key_to_eascii = {
+    SHIFT_KEY_TO_EASCII = {
         pygame.K_F1: uea.SHIFT_F1,
         pygame.K_F2: uea.SHIFT_F2,
         pygame.K_F3: uea.SHIFT_F3,
@@ -859,7 +913,7 @@ if pygame:
         pygame.K_KP5: uea.SHIFT_KP5,
     }
 
-    ctrl_key_to_eascii = {
+    CTRL_KEY_TO_EASCII = {
         pygame.K_F1: uea.CTRL_F1,
         pygame.K_F2: uea.CTRL_F2,
         pygame.K_F3: uea.CTRL_F3,
@@ -890,7 +944,7 @@ if pygame:
         pygame.K_MINUS: uea.CTRL_MINUS,
     }
 
-    alt_key_to_eascii = {
+    ALT_KEY_TO_EASCII = {
         pygame.K_1: uea.ALT_1,
         pygame.K_2: uea.ALT_2,
         pygame.K_3: uea.ALT_3,
@@ -950,7 +1004,7 @@ if pygame:
         pygame.K_KP5: uea.ALT_KP5,
     }
 
-    mod_to_scan = {
+    MOD_TO_SCAN = {
         pygame.KMOD_LSHIFT: scancode.LSHIFT,
         pygame.KMOD_RSHIFT: scancode.RSHIFT,
         pygame.KMOD_LCTRL: scancode.CTRL,
@@ -960,7 +1014,7 @@ if pygame:
         #pygame.KMOD_RALT: scancode.ALT,
     }
 
-    key_to_eascii_num = {
+    KEY_TO_EASCII_NUM = {
         pygame.K_KP0: u'0',
         pygame.K_KP1: u'1',
         pygame.K_KP2: u'2',
@@ -974,11 +1028,15 @@ if pygame:
     }
 
 
-def apply_composite_artifacts(screen, pixels=4):
+def apply_composite_artifacts(screen, bpp_out, bpp_in):
     """Process the canvas to apply composite colour artifacts."""
     src_array = pygame.surfarray.array2d(screen)
-    return pygame.surfarray.make_surface(
-                    window.apply_composite_artifacts(src_array, pixels))
+    width, _ = src_array.shape
+    mask = 1<<bpp_in - 1
+    step = bpp_out // bpp_in
+    s = [(src_array[_p:width:step] & mask) << _p for _p in range(step)]
+    packed = numpy.repeat(numpy.array(s).sum(axis=0), step, axis=0)
+    return pygame.surfarray.make_surface(packed)
 
 
 def glyph_to_surface(glyph):
