@@ -1,8 +1,8 @@
 
 NAME = u'PC-BASIC'
-PACKAGE = b'pcbasic'
-VERSION = u'2.0.0'
-COPYRIGHT = u'(C) Copyright 2013--2018 Rob Hagemans.'
+PACKAGE = 'pcbasic'
+VERSION = u'2.0.2'
+COPYRIGHT = u'(C) Copyright 2013--2019 Rob Hagemans.'
 LICENCE = u'GPLv3'
 AUTHOR = u'Rob Hagemans'
 EMAIL = u'robhagemans@yahoo.co.uk'
@@ -23,5 +23,6 @@ CLASSIFIERS = [
     u'Topic :: Software Development :: Interpreters',
     u'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
     u'Programming Language :: Python :: 2.7',
+    u'Programming Language :: Python :: 3.5',
 ]
 KEYWORDS = [u'emulator', u'interpreter', u'basic', u'retro', u'legacy', u'gwbasic', u'basica', u'pcjr', u'tandy']
